from __future__ import annotations

from django.apps import apps as django_apps
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist

from edc_visit_tracking.constants import SCHEDULED

from ..constants import KEYED, NOT_REQUIRED, REQUIRED


class PersistantSingletonMixin:
    """A mixin to be declared with a `predicates` collection.

    Handles a singleton model that needs to be entered
    at some later visit if not at the intended visit.

    The `required` entry status of CRF metadata is moved to the
    last attended visit in the schedule until the CRF is submitted.
    """

    def persistant_singleton_required(
        self, visit, model=None, exclude_visit_codes: list[str] | None = None
    ) -> bool:
        """Returns True if the CRF model was not completed from any time
        after the exclude_visit_codes

        Updates metadata for other timepoints, except the last, to `not required`
        until the model is submitted.

        CRF model is a singleton.
        """
        model_cls = django_apps.get_model(model)
        try:
            obj = model_cls.objects.get(
                subject_visit__subject_identifier=visit.subject_identifier
            )
        except ObjectDoesNotExist:
            obj = None
            required = (
                True
                if visit == self.get_last_attended_scheduled_visit(visit)
                and visit.visit_code not in exclude_visit_codes
                else False
            )
        except MultipleObjectsReturned:
            # necessary if the collection schedule changes and a singleton form
            # is later allowed to be added as a PRN; that is, no longer a
            # singleton.
            obj = None
            required = False
        else:
            required = False
        if visit and visit.visit_code not in exclude_visit_codes:
            last_attended_visit = self.get_last_attended_scheduled_visit(visit)
            if last_attended_visit:
                self.set_other_crf_metadata_not_required(
                    obj=obj,
                    model_cls=model_cls,
                    last_attended_visit=last_attended_visit,
                )
        return required

    @staticmethod
    def set_other_crf_metadata_not_required(
        obj=None, model_cls=None, last_attended_visit=None
    ):
        """Updates CRF metadata for the given model setting
        all but last visit to NOT_REQUIRED OR, if the model
        has been submitted, all others but the KEYED one as
        NOT_REQUIRED.

        Model must be a singleton
        """
        crf_metadata_model_cls = django_apps.get_model("edc_metadata.crfmetadata")
        model = f"{model_cls._meta.app_label}.{model_cls._meta.model_name}"

        opts = dict(
            subject_identifier=last_attended_visit.subject_identifier,
            model=model,
            visit_schedule_name=last_attended_visit.visit_schedule_name,
            schedule_name=last_attended_visit.schedule_name,
            visit_code_sequence=0,
        )
        if not obj:
            crf_metadata_model_cls.objects.filter(**opts).exclude(
                timepoint=last_attended_visit.appointment.timepoint
            ).update(entry_status=NOT_REQUIRED)
            crf_metadata_model_cls.objects.filter(
                timepoint=last_attended_visit.appointment.timepoint, **opts
            ).update(entry_status=REQUIRED)
        else:
            crf_metadata_model_cls.objects.filter(**opts).exclude(entry_status=KEYED).update(
                entry_status=NOT_REQUIRED
            )

    @staticmethod
    def get_last_attended_scheduled_visit(visit):
        """Returns the last attended `subject_visit` model instance given
        any existing `subject_visit` in the schedule.
        """
        return (
            visit.__class__.objects.filter(
                subject_identifier=visit.subject_identifier,
                reason=SCHEDULED,
                visit_schedule_name=visit.visit_schedule_name,
                schedule_name=visit.schedule_name,
                visit_code_sequence=0,
            )
            .order_by("appointment__timepoint")
            .last()
        )



# Generated by Django 5.2.3 on 2025-06-29 21:11

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0074_rename_stockout_stockavailability_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="historicalstockrequestitem",
            name="in_stock",
        ),
        migrations.RemoveField(
            model_name="historicalstockrequestitem",
            name="received",
        ),
        migrations.RemoveField(
            model_name="historicalstockrequestitem",
            name="received_datetime",
        ),
        migrations.RemoveField(
            model_name="stockrequestitem",
            name="in_stock",
        ),
        migrations.RemoveField(
            model_name="stockrequestitem",
            name="received",
        ),
        migrations.RemoveField(
            model_name="stockrequestitem",
            name="received_datetime",
        ),
    ]



# Generated by Django 4.2.11 on 2024-06-18 02:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_qareports", "0003_alter_qareportnote_options_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="qareportnote",
            name="status",
            field=models.CharField(
                choices=[("New", "New"), ("feedback", "Feedback")],
                default="New",
                max_length=25,
            ),
        ),
    ]



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass, field

from .get_languages import get_languages


class SiteDomainRequiredError(Exception):
    pass


class SiteCountryRequiredError(Exception):
    pass


@dataclass(order=True)
class SingleSite:
    site_id: int = field(compare=True)
    name: str
    domain: str
    _: KW_ONLY
    country: str = None
    country_code: str = field(default=None, repr=False)
    language_codes: list[str] = field(default_factory=list, repr=False)
    title: str | None = field(default=None, repr=False)
    languages: dict[str, str] = field(init=False, repr=False)
    description: str = field(init=False)

    def __post_init__(self):
        self.languages = get_languages(self.language_codes, self.site_id)
        self.description = (self.title or self.name).title()

    def __str__(self):
        return str(self.domain)



from django.core.exceptions import ObjectDoesNotExist
from django.db.models.signals import post_save
from django.dispatch import receiver

from edc_constants.constants import INCOMPLETE


@receiver(
    post_save,
    weak=False,
    dispatch_uid="update_subject_visit_document_status_on_pre_save",
)
def update_subject_visit_document_status_on_pre_save(
    instance, raw, using, update_fields, **kwargs
):
    """Updates the subject visit document status when an appointment
    is changed.

    Changing document status to incomplete forces the
    user to edit the subject visit on the dashboard
    before proceeding to the CRFs.
    """
    if not raw:
        try:
            subject_visit = getattr(instance, instance.related_visit_model_attr())
        except ObjectDoesNotExist:
            pass
        except AttributeError:
            pass
        else:
            if instance._meta.label_lower == "edc_appointment.appointment":
                old_instance = instance.__class__.objects.get(id=instance.id)
                if old_instance.appt_datetime.date() != instance.appt_datetime.date():
                    if subject_visit.document_status != INCOMPLETE:
                        subject_visit.document_status = INCOMPLETE
                        subject_visit.save_base(update_fields=["document_status"])



from .model_notification import ModelNotification


class GradedEventNotification(ModelNotification):
    grade: int | None = None
    model: str | None = None
    create_fields: tuple[str] = ("ae_grade",)
    update_fields: tuple[str] = ("ae_grade",)

    def field_value_condition_on_create(self, field, current_value):
        return str(current_value) == str(self.grade)

    def field_value_condition_on_update(self, field, previous_value, current_value):
        """Returns True if the value has changed and matches self.grade"""
        return str(previous_value) != str(current_value) and str(current_value) == str(
            self.grade
        )



from .crf_button import CrfButton
from .go_to_forms_button import GotToFormsButton
from .requisition_button import RequisitionButton
from .subject_consent_dashboard_button import SubjectConsentDashboardButton
from .subject_consent_listboard_button import SubjectConsentListboardButton
from .subject_screening_button import SubjectScreeningButton
from .timepoint_status_button import TimepointStatusButton

__all__ = [
    "CrfButton",
    "GotToFormsButton",
    "RequisitionButton",
    "RequisitionButton",
    "SubjectConsentDashboardButton",
    "SubjectConsentListboardButton",
    "SubjectScreeningButton",
    "TimepointStatusButton",
]



# Generated by Django 4.1.2 on 2023-01-24 15:51

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0036_auto_20230124_1822"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="appointment",
            name="appt_type_old",
        ),
        migrations.RemoveField(
            model_name="historicalappointment",
            name="appt_type_old",
        ),
    ]



# Generated by Django 3.2.6 on 2021-09-09 23:34

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_navbar", "0002_auto_20210423_1451"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="navbar",
            options={"verbose_name": "Navbar"},
        ),
    ]



# Generated by Django 4.2.9 on 2024-01-30 05:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_auth", "0032_userprofile_is_multisite_viewer"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userprofile",
            name="is_multisite_viewer",
            field=models.BooleanField(
                default=False,
                help_text="Designates if this user may view data from sites other than the current.",
                verbose_name="Multisite viewer",
            ),
        ),
    ]



from __future__ import annotations

from dataclasses import dataclass, field
from uuid import UUID

__all__ = ["CrfButton"]

from edc_view_utils import DashboardModelButton


@dataclass
class CrfButton(DashboardModelButton):
    colors: tuple[str, str, str] = field(default=("warning", "success", "success"))

    @property
    def extra_kwargs(self) -> dict[str, str | int | UUID]:
        return {self.model_cls().related_visit_model_attr(): self.appointment.related_visit.id}

    @property
    def disabled(self) -> str:
        disabled = "disabled"
        if (
            (not self.model_obj and self.perms.add)
            or self.perms.change
            or (self.perms.view and self.model_obj)
        ):
            disabled = ""
        return disabled



from django.db import models

from edc_action_item.models.action_model_mixin import ActionModelMixin
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models import BaseUuidModel
from edc_sites.model_mixins import SiteModelMixin

from .constants import LTFU_ACTION
from .model_mixins import LtfuModelMixin


class Ltfu(
    NonUniqueSubjectIdentifierFieldMixin,
    LtfuModelMixin,
    SiteModelMixin,
    ActionModelMixin,
    BaseUuidModel,
):
    action_name = LTFU_ACTION

    class Meta(BaseUuidModel.Meta, NonUniqueSubjectIdentifierFieldMixin.Meta):
        verbose_name = "Loss to Follow Up"
        verbose_name_plural = "Loss to Follow Ups"
        indexes = (
            *NonUniqueSubjectIdentifierFieldMixin.Meta.indexes,
            *ActionModelMixin.Meta.indexes,
            *BaseUuidModel.Meta.indexes,
            models.Index(fields=["subject_identifier", "action_identifier", "site"]),
        )



from decimal import Decimal

from django import forms

from edc_reportable import MILLIMOLES_PER_LITER, ConversionNotHandled
from edc_reportable.utils import convert_units
from edc_utils.round_up import round_half_away_from_zero

from .constants import GLUCOSE_HIGH_READING


def validate_glucose_as_millimoles_per_liter(
    prefix: str, cleaned_data: dict
) -> None | Decimal:
    converted_value = None
    min_val = Decimal("0.00")
    max_val = Decimal("30.00")
    high_value = Decimal(f"{GLUCOSE_HIGH_READING}")
    if (value := cleaned_data.get(f"{prefix}_value")) and (
        units := cleaned_data.get(f"{prefix}_units")
    ):
        try:
            converted_value = convert_units(
                label=prefix,
                value=value,
                units_from=units,
                units_to=MILLIMOLES_PER_LITER,
            )
        except ConversionNotHandled as e:
            raise forms.ValidationError({f"{prefix}_units": str(e)}) from e
        if (
            not (min_val <= round_half_away_from_zero(converted_value, 2) <= max_val)
            and round_half_away_from_zero(converted_value, 2) != high_value
        ):
            raise forms.ValidationError(
                {
                    f"{prefix}_value": (
                        f"This value is out-of-range. Got {converted_value} mmol/L"
                    )
                }
            )
    return converted_value



from edc_constants.constants import FEMALE, MALE

from ...adult_age_options import adult_age_options
from ...formula import Formula
from ...units import (
    CELLS_PER_MILLIMETER_CUBED,
    EGFR_UNITS,
    FEMTOLITERS_PER_CELL,
    GRAMS_PER_DECILITER,
    GRAMS_PER_LITER,
    IU_LITER,
    MICROMOLES_PER_LITER,
    MILLIGRAMS_PER_DECILITER,
    MILLIGRAMS_PER_LITER,
    MILLIMOLES_PER_LITER,
    PERCENT,
    PICOGRAMS_PER_CELL,
    PLUS,
    TEN_X_9_PER_LITER,
)

__all__ = ["normal_data"]

normal_data = {
    "albumin": [
        Formula(
            "3.5<=x<=5.0",
            units=GRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "35<=x<=50",
            units=GRAMS_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "alp": [Formula("40<=x<=150", units=IU_LITER, gender=[MALE, FEMALE], **adult_age_options)],
    "alt": [
        Formula("0<=x<=40", units=IU_LITER, gender=[MALE], **adult_age_options),
        Formula("0<=x<=35", units=IU_LITER, gender=[FEMALE], **adult_age_options),
    ],
    "amylase": [
        Formula("25<=x<=125", units=IU_LITER, gender=[MALE, FEMALE], **adult_age_options)
    ],
    "ast": [Formula("5<=x<=34", units=IU_LITER, gender=[MALE, FEMALE], **adult_age_options)],
    "egfr": [
        Formula(
            "0.0<=x<45.0",
            units=EGFR_UNITS,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "egfr_drop": [
        Formula(
            "x<40.0",
            units=PERCENT,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "chol": [
        Formula(
            "0.5<=x<=6.2",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "creatinine": [
        Formula(
            "63.6<=x<=110.5",
            units=MICROMOLES_PER_LITER,
            gender=[MALE],
            **adult_age_options,
        ),
        Formula(
            "50.4<=x<=98.1",
            units=MICROMOLES_PER_LITER,
            gender=[FEMALE],
            **adult_age_options,
        ),
        Formula(
            "0.6<=x<=1.3",
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "crp": [
        Formula(
            "0.0<=x<=0.5",
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "0.0<=x<=5.0",
            units=MILLIGRAMS_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "hdl": [
        Formula(
            "1.04<=x<=1.55",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "ggt": [
        Formula("12<=x<=64", units=IU_LITER, gender=[MALE], **adult_age_options),
        Formula("9<=x<=36", units=IU_LITER, gender=[FEMALE], **adult_age_options),
    ],
    "haemoglobin": [
        Formula(
            "13.0<=x<=17.0",
            units=GRAMS_PER_DECILITER,
            gender=[MALE],
            **adult_age_options,
        ),
        Formula(
            "12.0<=x<=15.0",
            units=GRAMS_PER_DECILITER,
            gender=[FEMALE],
            **adult_age_options,
        ),
    ],
    "hba1c": [
        Formula("4.4<=x<6.5", units=PERCENT, gender=[MALE, FEMALE], **adult_age_options)
    ],
    # hematocrit
    "hct": [
        Formula("37.0<=x<=54.0", units=PERCENT, gender=[MALE, FEMALE], **adult_age_options)
    ],
    "ldl": [
        Formula(
            "0.00<=x<=3.34",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "magnesium": [
        Formula(
            "0.75<=x<=1.2",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "1.8<=x<=2.9",
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "mch": [
        Formula(
            "27.6<=x<=33.4",
            units=PICOGRAMS_PER_CELL,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "mchc": [
        Formula(
            "32.8<=x<35.4",
            units=GRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "mcv": [
        Formula(
            "81.6<=x<=99.8",
            units=FEMTOLITERS_PER_CELL,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "neutrophil": [
        Formula(
            "2.5<=x<=7.5",
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "platelets": [
        Formula(
            "150<=x<=450",
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "150000<=x<=450000",
            units=CELLS_PER_MILLIMETER_CUBED,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "potassium": [
        Formula(
            "3.6<=x<=5.2",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "proteinuria": [
        Formula(
            "0.0<=x<1.0",
            units=PLUS,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "sodium": [
        Formula(
            "136<=x<=145",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "trig": [
        Formula(
            "0.00<=x<=1.69",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    # BUN
    "urea": [
        Formula(
            "2.5<=x<=6.5",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "uric_acid": [
        Formula(
            "0.15<=x<=0.35",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "7.2<=x",
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "rbc": [
        Formula(
            "3.5<=x<=5.5",
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "3500<=x<=5500",
            units=CELLS_PER_MILLIMETER_CUBED,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "tbil": [
        Formula(
            "0.2923<=x<1.2278",
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "0.00292<=x<0.0123",
            units=GRAMS_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        Formula(
            "5.0<=x<21.0",
            units=MICROMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "wbc": [
        Formula(
            "2.49<x",
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
}



from django.utils.translation import gettext_lazy as _

from edc_constants.constants import (
    COMPLETE,
    EQ,
    GT,
    GTE,
    LT,
    LTE,
    NOT_APPLICABLE,
    OTHER,
    PARTIAL,
    PENDING,
)
from edc_metadata.constants import NOT_REQUIRED
from edc_reportable import (
    MICROMOLES_PER_LITER,
    MICROMOLES_PER_LITER_DISPLAY,
    MILLIGRAMS_PER_DECILITER,
    MILLIMOLES_PER_LITER,
    MILLIMOLES_PER_LITER_DISPLAY,
)

from .constants import FILL_ACROSS, FILL_DOWN, FINGER_PRICK, TUBE

ABS_CALC = (("absolute", "Absolute"), ("calculated", "Calculated"))

ALIQUOT_STATUS = (("available", "available"), ("consumed", "consumed"))

ALIQUOT_CONDITIONS = (
    ("10", _("OK")),
    ("20", _("Inadequate volume for testing")),
    ("30", _("Clotted or haemolised")),
    ("40", _("Wrong tube type, unable to test")),
    ("50", _("Sample degradation has occured. Unsuitable for testing")),
    ("60", _("Expired tube")),
    ("70", _("Technical problem at lab, unable to test")),
)

FILL_ORDER = ((FILL_ACROSS, _("Across")), (FILL_DOWN, _("Down")))

MODIFY_ACTIONS = (
    ("INSERT", _("Insert")),
    ("UPDATE", _("Update")),
    ("DELETE", _("Delete")),
    ("PRINT", _("Print")),
    ("VIEW", _("Print")),
)

ORDER_STATUS = (
    (PENDING, _("Pending")),
    (PARTIAL, _("Partial")),
    (COMPLETE, _("Complete")),
)

GLUCOSE_UNITS = (
    (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
    (MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER_DISPLAY),
)

GLUCOSE_UNITS_NA = (
    (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
    (MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER_DISPLAY),
    (NOT_APPLICABLE, _("Not applicable")),
)
SERUM_CREATININE_UNITS = (
    (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
    (MICROMOLES_PER_LITER, MICROMOLES_PER_LITER_DISPLAY),
)

SERUM_CREATININE_UNITS_NA = (
    (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
    (MICROMOLES_PER_LITER, MICROMOLES_PER_LITER_DISPLAY),
    (NOT_APPLICABLE, _("Not applicable")),
)

RESULT_RELEASE_STATUS = (
    ("NEW", _("New")),
    ("RELEASED", _("Released")),
    ("AMENDED", _("Amended")),
)

RESULT_VALIDATION_STATUS = (
    ("P", _("Preliminary")),
    ("F", _("Final")),
    ("R", _("Rejected")),
)

RESULT_QUANTIFIER = ((EQ, EQ), (GT, GT), (GTE, GTE), (LT, LT), (LTE, LTE))

RESULT_QUANTIFIER_NA = (
    (EQ, EQ),
    (GT, GT),
    (GTE, GTE),
    (LT, LT),
    (LTE, LTE),
    (NOT_APPLICABLE, _("Not applicable")),
)

VL_QUANTIFIER = (
    (EQ, EQ),
    (GT, GT),
    (LT, LT),
)

VL_QUANTIFIER_NA = (
    (EQ, EQ),
    (GT, GT),
    (LT, LT),
    (NOT_APPLICABLE, _("Not applicable")),
)


SPECIMEN_MEASURE_UNITS = (
    ("mL", "mL"),
    ("uL", "uL"),
    ("spots", _("spots")),
    ("n/a", _("Not applicable")),
)

SPECIMEN_MEDIUM = (
    ("tube_any", _("Tube")),
    ("tube_edta", _("Tube EDTA")),
    ("swab", _("Swab")),
    ("dbs_card", _("DBS Card")),
)

UNITS = (
    ("%", "%"),
    ("10^3/uL", "10^3/uL"),
    ("10^3uL", "10^3uL"),
    ("10^6/uL", "10^6/uL"),
    ("cells/ul", "cells/ul"),
    ("copies/ml", "copies/ml"),
    ("fL", "fL"),
    ("g/dL", "g/dL"),
    ("g/L", "g/L"),
    ("mg/L", "mg/L"),
    ("mm/H", "mm/H"),
    ("mmol/L", "mmol/L"),
    ("ng/ml", "ng/ml"),
    ("pg", "pg"),
    ("ratio", "ratio"),
    ("U/L", "U/L"),
    ("umol/L", "umol/L"),
)

PRIORITY = (("normal", _("Normal")), ("urgent", _("Urgent")))

REASON_NOT_DRAWN = (
    (NOT_APPLICABLE, _("Not applicable")),
    ("collection_failed", _("Tried, but unable to obtain sample from patient")),
    ("absent", _("Patient did not attend visit")),
    ("refused", _("Patient refused")),
    ("no_supplies", _("No supplies")),
    (NOT_REQUIRED, _("No longer required for this visit")),
    (OTHER, _("Other")),
)

ITEM_TYPE = (
    (NOT_APPLICABLE, _("Not applicable")),
    (TUBE, _("Tube")),
    (FINGER_PRICK, _("Finger prick")),
    ("swab", _("Swab")),
    ("dbs", _("DBS Card")),
    (OTHER, _("Other")),
)



from __future__ import annotations

from typing import TYPE_CHECKING
from warnings import warn

from edc_registration import get_registered_subject
from edc_registration.utils import RegisteredSubjectDoesNotExist

from ..exceptions import InvalidSiteForSubjectError
from .get_site_model_cls import get_site_model_cls

if TYPE_CHECKING:
    from django.contrib.sites.models import Site

    from edc_registration.models import RegisteredSubject

__all__ = ["valid_site_for_subject_or_raise"]


def valid_site_for_subject_or_raise(
    subject_identifier: str, skip_get_current_site: bool | None = None
) -> Site:
    """Raises an InvalidSiteError exception if the subject_identifier is not
    from the current site.

    * Confirms by querying RegisteredSubject.
    * If subject_identifier is invalid will raise ObjectDoesNotExist
    """
    registered_subject: RegisteredSubject | None = get_registered_subject(
        subject_identifier, raise_exception=True
    )
    if skip_get_current_site:
        warn("Skipping validation of current site against registered subject site.")
        site_obj = registered_subject.site
    else:
        site_obj: Site = get_site_model_cls().objects.get_current()
        try:
            get_registered_subject(
                subject_identifier,
                site=site_obj,
                raise_exception=True,
            )
        except RegisteredSubjectDoesNotExist:
            if not registered_subject.site_id:
                raise InvalidSiteForSubjectError(
                    "Site not defined for registered subject! "
                    f"Subject identifier=`{subject_identifier}`. "
                )
            raise InvalidSiteForSubjectError(
                f"Invalid site for subject. Subject identifier=`{subject_identifier}`. "
                f"Expected `{registered_subject.site.name}`. "
                f"Got site_id=`{site_obj.id}`"
            )
    return site_obj



# Generated by Django 4.2.1 on 2023-07-07 19:32

import django.db.models.manager
from django.db import migrations

import edc_sites.models


class Migration(migrations.Migration):
    dependencies = [
        (
            "edc_data_manager",
            "0028_alter_dataquery_options_alter_queryrule_options_and_more",
        ),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="dataquery",
            managers=[
                ("objects", django.db.models.manager.Manager()),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
    ]



from django import forms
from django.core.exceptions import ObjectDoesNotExist
from django.urls.base import reverse
from django.utils.html import format_html

from edc_constants.constants import OTHER
from edc_dashboard.url_names import url_names
from edc_form_validators import FormValidator, FormValidatorMixin
from edc_registration.models import RegisteredSubject
from edc_screening.utils import get_subject_screening_model_cls, is_eligible_or_raise

from .models import SubjectRefusal


class SubjectRefusalFormValidator(FormValidator):
    def clean(self):
        self.required_if(OTHER, field="reason", field_required="other_reason")


class ScreeningFormMixin:
    def clean(self):
        cleaned_data = super().clean()
        screening_identifier = (
            cleaned_data.get("screening_identifier") or self.instance.screening_identifier
        )
        if screening_identifier:
            subject_screening = get_subject_screening_model_cls().objects.get(
                screening_identifier=screening_identifier
            )
            is_eligible_or_raise(subject_screening=subject_screening)
        return cleaned_data


class AlreadyConsentedFormMixin:
    def clean(self):
        cleaned_data = super().clean()
        try:
            obj = RegisteredSubject.objects.get(
                screening_identifier=self.instance.screening_identifier
            )
        except ObjectDoesNotExist:
            pass
        else:
            url_name = url_names.get("subject_dashboard_url")
            url = reverse(
                url_name,
                kwargs={"subject_identifier": obj.subject_identifier},
            )
            msg = format_html(
                "Not allowed. Subject has already consented. "
                'See subject <A href="{}">{}</A>',
                url,
                obj.subject_identifier,
            )
            raise forms.ValidationError(msg)
        return cleaned_data


class SubjectRefusalForm(
    AlreadyConsentedFormMixin, ScreeningFormMixin, FormValidatorMixin, forms.ModelForm
):
    form_validator_cls = SubjectRefusalFormValidator

    screening_identifier = forms.CharField(
        label="Screening identifier",
        widget=forms.TextInput(attrs={"readonly": "readonly"}),
    )

    class Meta:
        model = SubjectRefusal
        fields = "__all__"



from django.core.management.base import BaseCommand

from edc_data_manager.models import QueryRule
from edc_data_manager.rule import update_query_rules


class Command(BaseCommand):
    help = "Run active data query rules"

    def add_arguments(self, parser):
        parser.add_argument(
            "--verbose",
            type=bool,
            help="Verbosity (default=True)",
            default=True,
        )

        parser.add_argument(
            "--title",
            type=str,
            help="Titles. Separate by comma. Default is all rules",
            default="",
        )

    def handle(self, *args, **options):
        opts = dict(verbose=options["verbose"])
        if options["title"]:
            opts.update(
                pks=[
                    o.pk
                    for o in QueryRule.objects.filter(
                        title__in=options["title"].strip().split(",")
                    )
                ]
            )
        update_query_rules(**opts)



from edc_auth.site_auths import site_auths

from .auth_objects import EDC_FACILITY, EDC_FACILITY_SUPER, EDC_FACILITY_VIEW, codenames


def update_site_auths():
    site_auths.add_group(*codenames, name=EDC_FACILITY_VIEW, view_only=True)
    site_auths.add_group(*codenames, name=EDC_FACILITY, no_delete=True)
    site_auths.add_group(*codenames, name=EDC_FACILITY_SUPER)


update_site_auths()



from string import Template

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.core.mail import EmailMessage
from mempass import PasswordGenerator

from edc_protocol.research_protocol_config import ResearchProtocolConfig


class PasswordSetterError(Exception):
    pass


class PasswordSetter:
    email_body_template = Template(
        "Hi $first_name, \n\n"
        "Your clinicedc.org password has been reset.\n\n"
        "Your new password is:\n\n$password\n\n"
        "Your password was reset by $administrator.\n\n"
        "Thanks.\n\n"
        f"{ResearchProtocolConfig().project_name}\n\n"
    )

    def __init__(
        self,
        super_username: str,
        alternate_email: str | None = None,
        nwords: int | None = None,
        **kwargs,
    ) -> None:
        self.administrator_fullname = self.get_administrator_fullname(super_username)
        self.alternate_email = alternate_email
        self.password_generator = PasswordGenerator(nwords=nwords, **kwargs)

    @staticmethod
    def get_administrator_fullname(super_username: str) -> str:
        try:
            obj = User.objects.get(username=super_username)
        except ObjectDoesNotExist:
            raise PasswordSetterError(
                "Need the username of the administrator / super user. Got None."
            )
        return f"{obj.first_name} {obj.last_name}"

    def reset_all(self) -> None:
        users = User.objects.filter(is_active=True, is_staff=True, is_superuser=False)
        self._reset(users)

    def reset_by_groups(self, group_names: list[str] = None) -> None:
        users = User.objects.filter(
            groups__name__in=group_names,
            is_active=True,
            is_staff=True,
            is_superuser=False,
        )
        self._reset(users)

    def reset_users(self, usernames: list[str]) -> None:
        users = User.objects.filter(username__in=usernames, is_active=True, is_staff=True)
        self._reset(users)

    def reset_user(self, username: str) -> None:
        usernames = [username]
        self.reset_users(usernames)

    def reset_by_sites(self, site_names: list[str] = None) -> None:
        users = User.objects.filter(
            userprofile__sites__name__in=site_names,
            is_active=True,
            is_staff=True,
            is_superuser=False,
        )
        self._reset(users)

    def _reset(self, users: list[User]) -> None:
        for user in users:
            password = self.password_generator.get_password()
            user.set_password(password)
            user.save()
            body = self.email_body_template.safe_substitute(
                first_name=user.first_name,
                password=password,
                administrator=self.administrator_fullname,
            )
            email = EmailMessage(
                subject=(
                    f"{ResearchProtocolConfig().project_name}: "
                    "Your clinicedc.org password has been reset."
                ),
                body=body,
                from_email="noreply@clinicedc.org",
                to=[self.alternate_email if self.alternate_email else user.email],
            )
            email.send()



# Generated by Django 6.0 on 2025-03-27 22:41

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0068_stockout"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalstoragebin",
            name="capacity",
            field=models.PositiveIntegerField(
                default=0, validators=[django.core.validators.MinValueValidator(0)]
            ),
        ),
        migrations.AlterField(
            model_name="historicalstoragebin",
            name="name",
            field=models.CharField(
                blank=True,
                db_index=True,
                help_text="May be left blank or set to any unique name.",
                max_length=10,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="storagebin",
            name="capacity",
            field=models.PositiveIntegerField(
                default=0, validators=[django.core.validators.MinValueValidator(0)]
            ),
        ),
        migrations.AlterField(
            model_name="storagebin",
            name="name",
            field=models.CharField(
                blank=True,
                help_text="May be left blank or set to any unique name.",
                max_length=10,
                null=True,
                unique=True,
            ),
        ),
    ]



import sys
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.management import color_style

from .data import molecular_weights
from .utils import load_all_reference_ranges

if TYPE_CHECKING:
    from .models import MolecularWeight

style = color_style()


def load_mw():
    model_cls: MolecularWeight = django_apps.get_model("edc_reportable", "MolecularWeight")
    for label, mw in molecular_weights.items():
        model_cls.objects.get_or_create(label=label, mw=mw)


def post_migrate_load_reference_ranges(sender=None, **kwargs):
    sys.stdout.write(style.MIGRATE_HEADING("Loading reference ranges (reportables):\n"))
    load_mw()
    load_all_reference_ranges()



from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypeVar
from uuid import UUID

from django.utils import timezone

from edc_consent import site_consents
from edc_pdutils.site import Site
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_view_utils import ModelButton

if TYPE_CHECKING:
    from edc_consent.model_mixins import ConsentModelMixin
    from edc_screening.model_mixins import ScreeningModelMixin

    ScreeningModel = TypeVar("ScreeningModel", bound=ScreeningModelMixin)
    ConsentModel = TypeVar("ConsentModel", bound=ConsentModelMixin)

__all__ = ["SubjectConsentListboardButton"]


@dataclass
class SubjectConsentListboardButton(ModelButton):
    """For the consent button on subject screening listboard.

    Note: model_cls comes from the consent definition. The consent
    definition is selected by date. If there are multiple consent
    versions using different consent model classes, the date used to
    select the consent definition might matter. This may be a problem
    if data is not collected in realtime. That is, if there is a
    duration between the screening report_datetime and the date the
    consent button is rendered and clicked that crosses over from
    consent version 1 to version 2.
    """

    screening_obj: ScreeningModel = field(default=None)
    model_obj: ConsentModel = field(default=None)
    model_cls: type[ConsentModel] = field(default=None)
    consent_version: str = field(default=None, init=False)

    def __post_init__(self):
        cdef = site_consents.get_consent_definition(
            report_datetime=timezone.now(),
            screening_model=self.screening_obj._meta.label_lower,
        )
        self.model_cls = cdef.model_cls
        self.consent_version = cdef.version
        if self.screening_obj.consented:
            self.model_obj = self.model_cls.objects.get(
                subject_identifier=self.screening_obj.subject_identifier
            )
        if not self.next_url_name:
            self.next_url_name = "screening_listboard_url"

    @property
    def site(self) -> Site | None:
        return getattr(self.screening_obj, "site", None) or getattr(self.request, "site", None)

    @property
    def label(self) -> str:
        return f"Consent v{self.consent_version}"

    @property
    def reverse_kwargs(self) -> dict[str, str | UUID]:
        kwargs = dict(screening_identifier=self.screening_obj.screening_identifier)
        if re.match(
            ResearchProtocolConfig().subject_identifier_pattern,
            self.screening_obj.subject_identifier,
        ):
            kwargs.update(subject_identifier=self.screening_obj.subject_identifier)
        return kwargs

    @property
    def extra_kwargs(self) -> dict[str, str | int]:
        return dict(
            gender=self.screening_obj.gender,
            initials=self.screening_obj.initials,
            site=self.screening_obj.site.id,
        )



from django.db.models.signals import post_save
from django.dispatch import receiver


@receiver(
    post_save,
    weak=False,
    dispatch_uid="update_registered_subject_from_model_on_post_save",
)
def update_registered_subject_from_model_on_post_save(
    sender, instance, raw, created, using, **kwargs
):
    """Updates RegisteredSubject from models using
    UpdatesOrCreatesRegistrationModelMixin.
    """
    if not raw and not kwargs.get("update_fields"):
        try:
            instance.registration_update_or_create()
        except AttributeError as e:
            if "registration_update_or_create" not in str(e):
                raise AttributeError(str(e))



# Generated by Django 3.2.13 on 2022-09-06 23:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0013_auto_20220907_0202"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalrxrefill",
            name="refill_end_datetime",
            field=models.DateTimeField(
                help_text="Ending date for this refill",
                null=True,
                verbose_name="Refill end date/time",
            ),
        ),
        migrations.AddField(
            model_name="rxrefill",
            name="refill_end_datetime",
            field=models.DateTimeField(
                help_text="Ending date for this refill",
                null=True,
                verbose_name="Refill end date/time",
            ),
        ),
    ]



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

from django.core.exceptions import ObjectDoesNotExist

from .reference_range_colllection_model_cls import reference_range_colllection_model_cls

if TYPE_CHECKING:
    from ..models import ReferenceRangeCollection


__all__ = ["get_reference_range_collection"]


def get_reference_range_collection(obj) -> ReferenceRangeCollection:
    """Returns the reference range collection instance"""
    reference_range_collection = None
    if obj:
        try:
            name = obj.requisition.panel_object.reference_range_collection_name
        except AttributeError:
            name = obj.panel_object.reference_range_collection_name
        with contextlib.suppress(ObjectDoesNotExist):
            reference_range_collection = reference_range_colllection_model_cls().objects.get(
                name=name
            )
    return reference_range_collection



from django.db import models


class RequisitionStatusMixin(models.Model):
    received = models.BooleanField(default=False)

    received_datetime = models.DateTimeField(null=True, blank=True)

    processed = models.BooleanField(default=False)

    processed_datetime = models.DateTimeField(null=True, blank=True)

    packed = models.BooleanField(default=False)

    packed_datetime = models.DateTimeField(null=True, blank=True)

    shipped = models.BooleanField(default=False)

    shipped_datetime = models.DateTimeField(null=True, blank=True)

    class Meta:
        abstract = True



class RequisitionFormMixin:
    default_item_type = "tube"
    default_item_count = 1
    default_estimated_volume = 5.0

    # TODO: This is set by ModelAdmin get_initial_data. Should be removed.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not kwargs.get("instance"):
            if self.fields.get("item_type"):
                self.fields["item_type"].initial = self.default_item_type
            if self.fields.get("item_count"):
                self.fields["item_count"].initial = self.default_item_count
            if self.fields.get("estimated_volume"):
                self.fields["estimated_volume"].initial = self.default_estimated_volume
        if self.fields.get("specimen_type"):
            self.fields["specimen_type"].widget.attrs["readonly"] = True



from django.contrib import admin

from ..admin_site import edc_export_admin
from ..models import ObjectHistory


@admin.register(ObjectHistory, site=edc_export_admin)
class ObjectHistoryAdmin(admin.ModelAdmin):
    date_hierarchy = "created"
    list_display = (
        "export_uuid",
        "timestamp",
        "render",
        "status",
        "model",
        "export_change_type",
        "exported",
        "received",
        "received_datetime",
        "created",
    )
    list_filter = (
        "status",
        "exported",
        "received",
        "model",
        "export_change_type",
        "received_datetime",
        "created",
    )
    search_fields = ("export_uuid", "tx_pk", "tx")



from edc_navbar import Navbar, NavbarItem, site_navbars

pharmacy_navbar_item = NavbarItem(
    name="pharmacy",
    label="",
    title="",
    fa_icon="fa-prescription",
    codename="edc_pharmacy.nav_pharmacy_section",
    url_name="edc_pharmacy:home_url",
)

navbar = Navbar(name="pharmacy")

navbar.register(pharmacy_navbar_item)
site_navbars.register(navbar)



# Generated by Django 2.0b1 on 2017-11-16 14:06

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("edc_lab", "0008_auto_20170921_0719")]

    operations = [
        migrations.AlterModelOptions(
            name="historicalaliquot",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical ",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalbox",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical ",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalboxitem",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical ",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalconsignee",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical ",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalmanifest",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical ",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalshipper",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical ",
            },
        ),
    ]



# Generated by Django 5.2.6 on 2025-09-17 16:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_protocol_incident",
            "0023_alter_historicalprotocoldeviationviolation_revision_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="actionsrequired",
            name="extra_value",
            field=models.CharField(default="", max_length=250),
        ),
        migrations.AlterField(
            model_name="actionsrequired",
            name="field_name",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                help_text="Not required",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="actionsrequired",
            name="plural_name",
            field=models.CharField(
                default="", max_length=250, verbose_name="Plural name"
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="action_identifier",
            field=models.CharField(
                blank=True, db_index=True, default="", max_length=50
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="action_item_reason",
            field=models.TextField(default="", editable=False),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="corrective_action",
            field=models.TextField(
                blank=True, default="", verbose_name="Corrective action taken"
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to parent reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="preventative_action",
            field=models.TextField(
                blank=True, default="", verbose_name="Preventative action taken"
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to related reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="safety_impact_details",
            field=models.TextField(
                blank=True, default="", verbose_name='If "Yes", provide details'
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="short_description",
            field=models.CharField(
                default="",
                help_text='Max 35 characters. Note: If this occurrence is a "violation" there is additional space below for a more detailed description',
                max_length=35,
                verbose_name="Provide a short description of this occurrence",
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="study_outcomes_impact_details",
            field=models.TextField(
                blank=True, default="", verbose_name='If "Yes", provide details'
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="violation_description",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Describe in full. Explain how the violation happened, what occurred, etc.",
                verbose_name="Describe the violation",
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="violation_other",
            field=models.CharField(
                blank=True,
                default="",
                max_length=75,
                verbose_name="If other, please specify",
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="violation_reason",
            field=models.TextField(
                blank=True,
                default="",
                verbose_name="Explain the reason why the violation occurred",
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="action_identifier",
            field=models.CharField(
                blank=True, db_index=True, default="", max_length=50
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="action_item_reason",
            field=models.TextField(default="", editable=False),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="corrective_action",
            field=models.TextField(
                blank=True, default="", verbose_name="Corrective action taken"
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="incident_description",
            field=models.TextField(
                default="",
                help_text="Describe in full. Explain how the incident happened, what occurred, etc.",
                verbose_name="Describe the incident",
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="incident_other",
            field=models.CharField(
                blank=True,
                default="",
                max_length=75,
                verbose_name="If other, please specify",
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="incident_reason",
            field=models.TextField(
                default="", verbose_name="Explain the reason why the incident occurred"
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to parent reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="preventative_action",
            field=models.TextField(
                blank=True, default="", verbose_name="Preventative action taken"
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="reasons_withdrawn",
            field=models.TextField(blank=True, default=""),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to related reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="safety_impact_details",
            field=models.TextField(
                blank=True, default="", verbose_name='If "Yes", provide details'
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="short_description",
            field=models.CharField(
                default="",
                help_text="Max 35 characters. Note: there is additional space below for a more detailed description",
                max_length=35,
                verbose_name="Provide a short description of this incident",
            ),
        ),
        migrations.AlterField(
            model_name="historicalprotocolincident",
            name="study_outcomes_impact_details",
            field=models.TextField(
                blank=True, default="", verbose_name='If "Yes", provide details'
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="action_identifier",
            field=models.CharField(blank=True, default="", max_length=50, unique=True),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="action_item_reason",
            field=models.TextField(default="", editable=False),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="corrective_action",
            field=models.TextField(
                blank=True, default="", verbose_name="Corrective action taken"
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to parent reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="preventative_action",
            field=models.TextField(
                blank=True, default="", verbose_name="Preventative action taken"
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to related reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="safety_impact_details",
            field=models.TextField(
                blank=True, default="", verbose_name='If "Yes", provide details'
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="short_description",
            field=models.CharField(
                default="",
                help_text='Max 35 characters. Note: If this occurrence is a "violation" there is additional space below for a more detailed description',
                max_length=35,
                verbose_name="Provide a short description of this occurrence",
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="study_outcomes_impact_details",
            field=models.TextField(
                blank=True, default="", verbose_name='If "Yes", provide details'
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="violation_description",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Describe in full. Explain how the violation happened, what occurred, etc.",
                verbose_name="Describe the violation",
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="violation_other",
            field=models.CharField(
                blank=True,
                default="",
                max_length=75,
                verbose_name="If other, please specify",
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="violation_reason",
            field=models.TextField(
                blank=True,
                default="",
                verbose_name="Explain the reason why the violation occurred",
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="action_identifier",
            field=models.CharField(blank=True, default="", max_length=50, unique=True),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="action_item_reason",
            field=models.TextField(default="", editable=False),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="corrective_action",
            field=models.TextField(
                blank=True, default="", verbose_name="Corrective action taken"
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="incident_description",
            field=models.TextField(
                default="",
                help_text="Describe in full. Explain how the incident happened, what occurred, etc.",
                verbose_name="Describe the incident",
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="incident_other",
            field=models.CharField(
                blank=True,
                default="",
                max_length=75,
                verbose_name="If other, please specify",
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="incident_reason",
            field=models.TextField(
                default="", verbose_name="Explain the reason why the incident occurred"
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to parent reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="preventative_action",
            field=models.TextField(
                blank=True, default="", verbose_name="Preventative action taken"
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="reasons_withdrawn",
            field=models.TextField(blank=True, default=""),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to related reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="safety_impact_details",
            field=models.TextField(
                blank=True, default="", verbose_name='If "Yes", provide details'
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="short_description",
            field=models.CharField(
                default="",
                help_text="Max 35 characters. Note: there is additional space below for a more detailed description",
                max_length=35,
                verbose_name="Provide a short description of this incident",
            ),
        ),
        migrations.AlterField(
            model_name="protocolincident",
            name="study_outcomes_impact_details",
            field=models.TextField(
                blank=True, default="", verbose_name='If "Yes", provide details'
            ),
        ),
        migrations.AlterField(
            model_name="protocolincidents",
            name="extra_value",
            field=models.CharField(default="", max_length=250),
        ),
        migrations.AlterField(
            model_name="protocolincidents",
            name="field_name",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                help_text="Not required",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="protocolincidents",
            name="plural_name",
            field=models.CharField(
                default="", max_length=250, verbose_name="Plural name"
            ),
        ),
        migrations.AlterField(
            model_name="protocolviolations",
            name="extra_value",
            field=models.CharField(default="", max_length=250),
        ),
        migrations.AlterField(
            model_name="protocolviolations",
            name="field_name",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                help_text="Not required",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="protocolviolations",
            name="plural_name",
            field=models.CharField(
                default="", max_length=250, verbose_name="Plural name"
            ),
        ),
    ]



from django.apps import apps as django_apps
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist


class DataDictionaryError(Exception):
    pass


def get_longitudinal_value(subject_identifier=None, reference_dt=None, model=None, field=None):
    DataDictionary = django_apps.get_model("edc_data_manager.datadictionary")
    opts = dict(field_name=field)
    if model:
        opts.update(model=model)
    try:
        data_dictionary = DataDictionary.objects.get(**opts)
    except ObjectDoesNotExist:
        raise DataDictionaryError(
            "Unable to determine longitudinal value. "
            f"Field name '{field}' not found in the data dictionary"
        )
    except MultipleObjectsReturned:
        fields = [
            f"{obj.model}.{obj.field_name}" for obj in DataDictionary.objects.filter(**opts)
        ]
        raise DataDictionaryError(
            "Unable to determine longitudinal value. "
            f"Field name '{field}' is ambiguous. Try specifying the model as well. "
            f"Got {fields}"
        )
    qs = data_dictionary.model_cls.objects.filter(
        subject_visit__appointment__subject_identifier=subject_identifier,
        report_datetime__lte=reference_dt,
    ).order_by("-report_datetime")
    if qs:
        return getattr(qs[0], field)
    return None



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from django_audit_fields import audit_fieldset_tuple
from rangefilter.filters import DateRangeFilterBuilder

from edc_model_admin.history import SimpleHistoryAdmin
from edc_sites.admin import SiteModelAdminMixin
from edc_utils.date import to_local

from ...admin_site import edc_pharmacy_admin
from ...models import ConfirmationAtSiteItem
from ..model_admin_mixin import ModelAdminMixin


@admin.register(ConfirmationAtSiteItem, site=edc_pharmacy_admin)
class ConfirmationAtSiteItemAdmin(SiteModelAdminMixin, ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Stock items confirmed at site"
    change_form_title = "Pharmacy: Stock item confirmed at site"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    actions = ["delete_selected"]

    ordering = ("-transfer_confirmation_item_identifier",)

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "transfer_confirmation_item_identifier",
                    "transfer_confirmation_item_datetime",
                    "confirmation_at_site",
                    "stock",
                    "confirmed_datetime",
                    "confirmed_by",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "transfer_confirmation_item_date",
        "subject",
        "site",
        "stock_changelist",
        "confirmation_at_site_changelist",
        "confirmed_datetime",
        "confirmed_by",
    )

    list_filter = (("transfer_confirmation_item_datetime", DateRangeFilterBuilder()),)

    readonly_fields = (
        "transfer_confirmation_item_identifier",
        "transfer_confirmation_item_datetime",
        "confirmation_at_site",
        "stock",
        "confirmed_datetime",
        "confirmed_by",
    )

    search_fields = (
        "pk",
        "confirmation_at_site__pk",
        "stock__code",
        "stock__pk",
        "stock__allocation__registered_subject__subject_identifier",
    )

    @admin.display(
        description="CONFIRMATION #", ordering="-transfer_confirmation_item_identifier"
    )
    def identifier(self, obj):
        return obj.transfer_confirmation_item_identifier

    @admin.display(description="Date", ordering="transfer_confirmation_item_datetime")
    def transfer_confirmation_item_date(self, obj):
        return to_local(obj.transfer_confirmation_item_datetime).date()

    @admin.display(description="Site", ordering="confirmation_at_site__location__site__id")
    def site(self, obj):
        return obj.confirmation_at_site.location.site.id

    @admin.display(
        description="SUBJECT #",
        ordering="stock__allocation__registered_subject__subject_identifier",
    )
    def subject(self, obj):
        return obj.stock.allocation.registered_subject.subject_identifier

    @admin.display(
        description="Transfer confirmation",
        ordering="confirmation_at_site__transfer_confirmation_identifier",
    )
    def confirmation_at_site_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_confirmationatsite_changelist")
        url = f"{url}?q={obj.confirmation_at_site.id}"
        context = dict(
            url=url,
            label=obj.confirmation_at_site.transfer_confirmation_identifier,
            title="Go to stock transfer confirmation",
        )
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Stock", ordering="stock__code")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stockproxy_changelist")
        url = f"{url}?q={obj.stock.code}"
        context = dict(url=url, label=obj.stock.code, title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    def get_view_only_site_ids_for_user(self, request) -> list[int]:
        return [s.id for s in request.user.userprofile.sites.all() if s.id != request.site.id]



import os
from warnings import warn

from django.conf import settings
from django.template.loader import select_template


class EdcTemplateDoesNotExist(Exception):
    pass


def get_index_page() -> int:
    index_page = getattr(settings, "INDEX_PAGE", None)
    if not index_page:
        warn("Settings attribute not set. See settings.INDEX_PAGE")
    return getattr(settings, "INDEX_PAGE", None)


def get_index_page_label() -> int:
    return getattr(settings, "INDEX_PAGE_LABEL", settings.APP_NAME)


def splitall(path):
    """Taken from
    https://www.oreilly.com/library/view/python-cookbook/0596001673/ch04s16.html
    """
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        if parts[1] == path:  # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        path = parts[0]
        allparts.insert(0, parts[1])
    return allparts


def select_edc_template(relative_path, default_app_label):
    """Returns a template object."""
    local_path = settings.APP_NAME
    default_path = default_app_label
    return select_template(
        [
            os.path.join(local_path, relative_path),
            os.path.join(default_path, relative_path),
        ]
    )


def get_dashboard_app_label():
    return getattr(settings, "EDC_DASHBOARD_APP_LABEL", None)



import csv
import uuid
from pathlib import Path

from django.apps import apps as django_apps
from django.core.exceptions import ValidationError
from django.utils import timezone

from ..utils import get_export_folder
from .file_history_updater import FileHistoryUpdater
from .object_history_helpers import ObjectHistoryHelper
from .value_getter import ValueGetter


class ModelExporterError(Exception):
    pass


class ModelExporterInvalidLookup(Exception):  # noqa: N818
    pass


class ModelExporterUnknownField(ValidationError):  # noqa: N818
    pass


class AdditionalValues:
    def __init__(self, export_datetime=None):
        self.export_uuid = str(uuid.uuid4())
        self.export_datetime = export_datetime
        self.timestamp = self.export_datetime.strftime("%Y%m%d%H%M%S")
        self.export_change_type = None


class ModelExporter:
    delimiter = "|"
    file_history_updater_cls = FileHistoryUpdater
    object_history_helper_cls = ObjectHistoryHelper
    value_getter_cls = ValueGetter
    additional_values_cls = AdditionalValues

    export_fields = (
        "export_uuid",
        "timestamp",
        "export_datetime",
        "export_change_type",
    )
    required_fields = ("subject_identifier", "report_datetime")
    audit_fields = (
        "hostname_created",
        "hostname_modified",
        "created",
        "modified",
        "user_created",
        "user_modified",
        "revision",
    )

    def __init__(
        self,
        queryset=None,
        model=None,
        field_names=None,
        exclude_field_names=None,
        lookups=None,
        exclude_m2m=None,
        encrypt=None,
    ):
        self._field_names = None
        self._model = model
        self._model_cls = None
        self.encrypt = True if encrypt is None else encrypt
        self.exclude_m2m = exclude_m2m
        self.lookups = lookups or {}
        self.queryset = queryset
        self.row = None
        self.row_instance = None
        self.exclude_field_names = exclude_field_names
        self.exclude_m2m = exclude_m2m

        self.field_names = field_names

    @property
    def field_names(self):
        return self._field_names

    @field_names.setter
    def field_names(self, field_names):
        if field_names:
            self._field_names = field_names
        else:
            self._field_names = [f.name for f in self.model_cls._meta.fields]
            if not self.exclude_m2m:
                for m2m in self.model_cls._meta.many_to_many:
                    self._field_names.append(m2m.name)
            if self.exclude_field_names:
                for f in self._field_names:
                    if f in self.exclude_field_names:
                        self.field_names.pop(self._field_names.index(f))

        for f in self._field_names:
            if f in self.export_fields or f in self.audit_fields or f in self.required_fields:
                self._field_names.pop(self._field_names.index(f))
        self._field_names = (
            *self.export_fields,
            *self.required_fields,
            *self.field_names,
            *self.audit_fields,
        )

    @property
    def model_cls(self):
        if not self._model_cls:
            try:
                self.queryset.count()
            except AttributeError:
                self._model_cls = django_apps.get_model(self._model)
            else:
                self._model_cls = self.queryset.model
        return self._model_cls

    def export(self, queryset=None) -> Path:
        """Writes the export file and returns the file name."""
        self.queryset = queryset or self.queryset
        exported_datetime = timezone.now()
        filename = self.get_filename(exported_datetime)
        path = get_export_folder() / filename
        with path.open("w") as f:
            csv_writer = csv.DictWriter(
                f, fieldnames=self.field_names, delimiter=self.delimiter
            )
            csv_writer.writeheader()
            for model_obj in self.queryset:
                object_helper = self.object_history_helper_cls(
                    model_obj=model_obj, create=True
                )
                objects = object_helper.get_not_exported()
                for obj in objects:
                    row = self.prepare_row(
                        model_obj=model_obj,
                        exported_datetime=exported_datetime,
                        export_change_type=obj.export_change_type,
                    )
                    csv_writer.writerow(row)
                object_helper.update_as_exported(
                    objects=objects, exported_datetime=exported_datetime
                )
        file_history_updater = self.file_history_updater_cls(
            path=path,
            delimiter=self.delimiter,
            model=self.model_cls._meta.label_lower,
            filename=filename,
        )
        file_history_updater.update()
        return path

    def prepare_row(self, model_obj=None, exported_datetime=None, export_change_type=None):
        """Returns one row for the CSV writer.

        Most of the work is done by the ValueGetter class.
        """
        additional_values = self.additional_values_cls(export_datetime=exported_datetime)
        row = {}
        for field_name in self.field_names:
            value_getter = self.value_getter_cls(
                field_name=field_name,
                model_obj=model_obj,
                additional_values=additional_values,
                lookups=self.lookups,
                encrypt=self.encrypt,
            )
            row.update({field_name: value_getter.value})
            row["exported"] = True
            row["exported_datetime"] = exported_datetime
            row["timestamp"] = exported_datetime.strftime("%Y%m%d%H%M%S")
            row["export_uuid"] = model_obj.export_uuid
            row["export_change_type"] = export_change_type
        return row

    def get_filename(self, exported_datetime=None):
        """Returns a CSV filename using the model label lower
        and the exported date.
        """
        formatted_model = self.model_cls._meta.label_lower.replace(".", "_")
        formatted_date = exported_datetime.strftime("%Y%m%d%H%M%S")
        return f"{formatted_model}_{formatted_date}.csv"



from __future__ import annotations

from typing import TYPE_CHECKING

from .history_button import HistoryButton
from .query_button import QueryButton

if TYPE_CHECKING:
    from edc_appointment.models import Appointment
    from edc_metadata.models import CrfMetadata
    from edc_registration.models import RegisteredSubject
    from edc_visit_schedule.models import VisitSchedule

__all__ = ["render_history_and_query_buttons"]


def render_history_and_query_buttons(
    context,
    model_obj: CrfMetadata = None,
    appointment: Appointment = None,
    registered_subject: RegisteredSubject = None,
    visit_schedule: VisitSchedule = None,
) -> tuple[HistoryButton, QueryButton]:
    # if still using deprecated ModelWrapper, get model instance
    # from model_wrapper
    appointment = getattr(appointment, "object", appointment)
    query_btn = QueryButton(
        metadata_model_obj=model_obj,
        user=context["request"].user,
        current_site=context["request"].site,
        appointment=appointment,
        registered_subject=registered_subject,
        visit_schedule=visit_schedule,
    )
    history_btn = HistoryButton(model_obj=model_obj)
    return history_btn, query_btn



from django.db.models import Q

from edc_constants.constants import ABNORMAL
from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin
from edc_screening.utils import get_subject_screening_model

from ...view_mixins import ListboardFilterViewMixin, SearchFormViewMixin
from ..listboard_view import ListboardView
from .listboard_view_filter import ScreeningListboardViewFilters


class ScreeningListboardView(
    EdcViewMixin,
    NavbarViewMixin,
    ListboardFilterViewMixin,
    SearchFormViewMixin,
    ListboardView,
):
    listboard_model = get_subject_screening_model()
    listboard_view_filters = ScreeningListboardViewFilters()

    listboard_template = "screening_listboard_template"
    listboard_url = "screening_listboard_url"
    listboard_panel_style = "info"
    listboard_fa_icon = "fas fa-user-plus"
    listboard_view_permission_codename = "edc_screening.view_screening_listboard"
    alternate_search_attr = "screening_identifier"
    navbar_selected_item = "screened_subject"
    ordering = "-report_datetime"
    paginate_by = 10
    search_form_url = "screening_listboard_url"
    search_fields = [
        "screening_identifier",
        "initials__exact",
        "subject_identifier",
        "user_created",
        "user_modified",
    ]

    def get_context_data(self, **kwargs) -> dict:
        kwargs.update(
            subject_screening_add_url=self.get_subject_screening_add_url(),
            ABNORMAL=ABNORMAL,
        )
        return super().get_context_data(**kwargs)

    def get_subject_screening_add_url(self) -> str:
        return self.listboard_model_cls().get_absolute_url()

    def get_queryset_filter_options(self, request, *args, **kwargs) -> tuple[Q, dict]:
        q_object, options = super().get_queryset_filter_options(request, *args, **kwargs)
        if kwargs.get("screening_identifier"):
            options.update({"screening_identifier": kwargs.get("screening_identifier")})
        return q_object, options



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from edc_model_admin.history import SimpleHistoryAdmin

from ...admin_site import edc_pharmacy_admin
from ...forms import SupplierForm
from ...models import Supplier
from ..model_admin_mixin import ModelAdminMixin


@admin.register(Supplier, site=edc_pharmacy_admin)
class SupplierAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Suppliers"
    change_form_title = "Pharmacy: Supplier"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True

    form = SupplierForm
    ordering = ("name",)

    fieldsets = (
        (
            None,
            {"fields": (["name", "contact"])},
        ),
        (
            "Address",
            {
                "fields": (
                    [
                        "address_one",
                        "address_two",
                        "city",
                        "postal_code",
                        "state",
                        "country",
                    ]
                )
            },
        ),
        (
            "Contact",
            {
                "fields": (
                    [
                        "email",
                        "email_alternative",
                        "telephone",
                        "telephone_alternative",
                    ]
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "name",
        "contact",
        "created",
        "modified",
    )
    list_filter = ("country",)
    search_fields = ("id", "supplier_identifier", "name", "contact")

    @admin.display(description="SUPPLIER #", ordering="-supplier_identifier")
    def identifier(self, obj):
        return obj.supplier_identifier



from .appointment_fields_model_mixin import AppointmentFieldsModelMixin
from .appointment_methods_model_mixin import (
    AppointmentMethodsModelError,
    AppointmentMethodsModelMixin,
)
from .appointment_model_mixin import AppointmentModelMixin
from .missed_appointment_model_mixin import MissedAppointmentModelMixin
from .next_appointment_crf_model_mixin import NextAppointmentCrfModelMixin
from .window_period_model_mixin import WindowPeriodModelMixin



from django.contrib.admin import display
from django.template.loader import render_to_string


class PdfButtonModelAdminMixin:
    """A model admin mixin to add a PDF download button for
    the model's custom `CrfPdfReport`.

    Add "pdf_button" to the changelist's `list_display`
    """

    pdf_button_url_name: str = "edc_pdf_reports:pdf_intermediate_url"
    pdf_button_title: str = "Download report as PDF"
    pdf_button_template_name: str = "edc_pdf_reports/pdf_button.html"

    @display(description="PDF")
    def pdf_button(self, obj):
        context = dict(
            str_pk=str(obj.id),
            app_label=obj._meta.app_label,
            model_name=obj._meta.model_name,
            title=self.pdf_button_title,
            url_name=self.pdf_button_url_name,
        )
        return render_to_string(template_name=self.pdf_button_template_name, context=context)



# Generated by Django 2.1.4 on 2019-01-31 18:04

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_action_item", "0015_auto_20190114_0250")]

    operations = [
        migrations.AlterField(
            model_name="historicalactionitem",
            name="parent_action_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="related_action_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_action_item.ActionItem",
            ),
        ),
    ]



from django.core.validators import MinValueValidator
from django.db import models
from django.db.models.deletion import PROTECT
from django.utils import timezone

from edc_action_item.managers import (
    ActionIdentifierModelManager,
    ActionIdentifierSiteManager,
)
from edc_action_item.models import ActionNoManagersModelMixin
from edc_constants.choices import YES_NO
from edc_identifier.model_mixins import UniqueSubjectIdentifierFieldMixin
from edc_model.models import HistoricalRecords
from edc_model.validators import date_not_future, datetime_not_future
from edc_model_fields.fields.other_charfield import OtherCharField
from edc_pdf_reports.model_mixins import PdfReportModelMixin
from edc_protocol.validators import (
    date_not_before_study_start,
    datetime_not_before_study_start,
)
from edc_sites.model_mixins import SiteModelMixin

from ...constants import DEATH_REPORT_ACTION
from ...models import CauseOfDeath
from ...pdf_reports import DeathPdfReport


class DeathReportModelMixin(
    SiteModelMixin,
    UniqueSubjectIdentifierFieldMixin,
    ActionNoManagersModelMixin,
    PdfReportModelMixin,
    models.Model,
):
    action_name = DEATH_REPORT_ACTION

    pdf_report_cls = DeathPdfReport

    death_date_field = "death_datetime"

    report_datetime = models.DateTimeField(
        verbose_name="Report Date",
        validators=[datetime_not_before_study_start, datetime_not_future],
        default=timezone.now,
    )

    death_datetime = models.DateTimeField(
        validators=[datetime_not_before_study_start, datetime_not_future],
        verbose_name="Date and Time of Death",
        null=True,
        blank=False,
    )

    death_date = models.DateField(
        validators=[date_not_before_study_start, date_not_future],
        verbose_name="Date of Death",
        null=True,
        blank=False,
    )

    study_day = models.IntegerField(
        validators=[MinValueValidator(1)],
        verbose_name="Study day",
        null=True,
        blank=False,
    )

    death_as_inpatient = models.CharField(
        choices=YES_NO,
        max_length=5,
        verbose_name="Death as inpatient",
        blank=False,
        default="",
    )

    cause_of_death = models.ForeignKey(
        CauseOfDeath,
        on_delete=PROTECT,
        verbose_name="Main cause of death",
        help_text=(
            "Main cause of death in the opinion of the local study doctor and local PI"
        ),
        null=True,
        blank=False,
    )

    cause_of_death_other = OtherCharField(max_length=100, blank=True, default="")

    narrative = models.TextField(verbose_name="Narrative", default="")

    objects = ActionIdentifierModelManager()

    on_site = ActionIdentifierSiteManager()

    history = HistoricalRecords(inherit=True)

    class Meta(
        SiteModelMixin.Meta,
        UniqueSubjectIdentifierFieldMixin.Meta,
        ActionNoManagersModelMixin.Meta,
    ):
        abstract = True
        verbose_name = "Death Report"
        verbose_name_plural = "Death Reports"
        indexes = (
            *ActionNoManagersModelMixin.Meta.indexes,
            models.Index(fields=["subject_identifier", "action_identifier", "site", "id"]),
        )

    def natural_key(self):
        return (self.action_identifier,)

    natural_key.dependencies = ("edc_adverse_event.causeofdeath",)



from __future__ import annotations

from typing import TYPE_CHECKING

from ..utils import get_duplicates
from .forms_collection import FormsCollection, FormsCollectionError

if TYPE_CHECKING:
    from .requisition import Requisition


class RequisitionCollection(FormsCollection):
    def __init__(self, *forms: Requisition, name: str | None = None, **kwargs):
        super().__init__(*forms, name=name, **kwargs)

    @staticmethod
    def collection_is_unique_or_raise(forms: tuple[Requisition]) -> None:
        panels = [f.name for f in forms if f.required]
        if duplicates := get_duplicates(list_items=panels):
            raise FormsCollectionError(
                "Expected be a unique sequence of requisitions/panels. "
                f"Got {sorted(panels)}. Duplicates {sorted(duplicates)}"
            )



from __future__ import annotations

from typing import TYPE_CHECKING

from django import forms
from django.apps import apps as django_apps

if TYPE_CHECKING:
    from django.contrib.sites.models import Site

__all__ = ["SiteModelFormMixin"]


class SiteModelFormMixin:
    """Validate the current site against a form question.

    This should be used sparingly. A good place is on the screening and/or consent form.

    Declare the modeladmin class with `SiteModelAdminMixin` to have
    the current site set on the form from the request object.

    Declare a `site` field with widget on the ModeForm:

    site = SiteField()

    You will also need to re-declare the `site` model field as `editable`.
    """

    def clean(self) -> dict:
        cleaned_data = super().clean()
        self.validate_with_current_site()
        return cleaned_data

    @property
    def site_model_cls(self) -> type[Site]:
        return django_apps.get_model("sites.site")

    @property
    def site(self) -> Site:
        if related_visit := getattr(self, "related_visit", None):
            return related_visit.site
        return (
            self.cleaned_data.get("site")
            or self.instance.site
            or self.site_model_cls.objects.get_current()
        )

    def validate_with_current_site(self) -> None:
        current_site = getattr(self, "current_site", None)
        if current_site and self.site and current_site.id != self.site.id:
            raise forms.ValidationError(
                {
                    "site": (
                        "Invalid. Please check you are logged into the correct site "
                        "before continuing"
                    )
                }
            )



from django.conf import settings
from edc_model_wrapper import ModelWrapper


class RequisitionModelWrapper(ModelWrapper):
    model = settings.SUBJECT_REQUISITION_MODEL
    next_url_name = "requisition_listboard_url"
    next_url_attrs = ["appointment", "subject_identifier"]
    querystring_attrs = ["subject_visit", "panel"]

    @property
    def subject_visit(self):
        return str(self.object.subject_visit.id)

    @property
    def appointment(self):
        return str(self.object.subject_visit.appointment.id)

    @property
    def subject_identifier(self):
        return self.object.subject_visit.subject_identifier

    @property
    def panel(self):
        return str(self.object.panel.id)



from .note_model_mixin import NoteModelMixin
from .on_study_missing_values_model_mixin import OnStudyMissingValuesModelMixin
from .qa_report_model_mixin import QaReportModelMixin
from .qa_reports_permissions import qa_reports_permissions



from __future__ import annotations

from typing import Any

from django.apps import apps as django_apps
from django.core.exceptions import PermissionDenied
from django.db.models import Q
from django.utils.translation import gettext as _
from django.views.generic.list import ListView

from edc_dashboard.view_mixins import (
    TemplateRequestContextMixin,
    UrlRequestContextMixin,
)
from edc_sites.site import sites
from edc_sites.view_mixins import SiteViewMixin

from ..view_mixins import QueryStringViewMixin, SearchListboardMixin


class ListboardViewError(Exception):
    pass


class BaseListboardView(SiteViewMixin, TemplateRequestContextMixin, ListView):
    listboard_model: str | None = None  # label_lower model name
    context_object_name: str = "results"

    empty_queryset_message: str = _("Nothing to display.")

    listboard_template: str | None = None  # an existing key in request.context_data

    # if self.listboard_url declared through another mixin.
    listboard_url: str | None = None  # an existing key in request.context_data
    listboard_back_url: str | None = None

    # styling
    # default, info, success, danger, warning, etc. See Bootstrap.
    listboard_panel_style: str = "default"
    listboard_fa_icon: str | None = None
    listboard_panel_title: str | None = None
    listboard_instructions: str | None = None
    show_change_form_button: bool = True

    # permissions
    permissions_warning_message: str = _("You do not have permission to view these data.")
    # e.g. "edc_subject_dashboard.view_subject_listboard"
    listboard_view_permission_codename: str | None = None
    # e.g. "edc_subject_dashboard.view_subject_listboard"
    listboard_view_only_my_permission_codename: str | None = None

    ordering: str = "-created"

    orphans: int = 3
    paginate_by: int = 10
    paginator_url = None  # defaults to listboard_url

    def get(self, request, *args, **kwargs):
        if not self.has_view_listboard_perms:
            raise PermissionDenied
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        if self.listboard_fa_icon and self.listboard_fa_icon.startswith("fa-"):
            self.listboard_fa_icon = f"fas {self.listboard_fa_icon}"
        kwargs.update(
            empty_queryset_message=self.get_empty_queryset_message(),
            listboard_fa_icon=self.listboard_fa_icon,
            listboard_panel_style=self.listboard_panel_style,
            listboard_panel_title=self.listboard_panel_title,
            listboard_instructions=self.listboard_instructions,
            show_change_form_button=self.show_change_form_button,
            # object_list=self.object_list,
            **self.add_url_to_context(
                new_key="listboard_url", existing_key=self.listboard_url
            ),
        )
        if self.listboard_back_url:
            kwargs.update(
                **self.add_url_to_context(
                    new_key="listboard_back_url",
                    existing_key=self.listboard_back_url,
                )
            )
        kwargs.update(
            has_listboard_model_perms=self.has_listboard_model_perms,
            has_view_listboard_perms=self.has_view_listboard_perms,
            listboard_view_permission_codename=self.listboard_view_permission_codename,
            permissions_warning_message=self.permissions_warning_message,
            **self.add_url_to_context(
                new_key="paginator_url",
                existing_key=self.paginator_url or self.listboard_url,
            ),
        )
        return super().get_context_data(**kwargs)

    def get_template_names(self):
        return [self.get_template_from_context(self.listboard_template)]

    @property
    def url_kwargs(self):
        """Returns a dictionary of URL options for either the
        Search form URL and the Form Action.
        """
        return {}

    def get_listboard_model(self) -> str:
        return self.listboard_model

    @property
    def listboard_model_cls(self):
        """Returns the listboard's model class.

        Accepts `listboard_model` as a model class or label_lower.
        """
        if not self.get_listboard_model():
            raise ListboardViewError(f"Listboard model not declared. Got None. See {self!r}")
        try:
            return django_apps.get_model(self.get_listboard_model())
        except (ValueError, AttributeError):
            return self.get_listboard_model()

    def get_queryset(self):
        """Return the queryset for this view.

        Completely overrides ListView.get_queryset.

        The returned queryset is set to self.object_list in the
        parent call to `get()` just before rendering to response.

        Note:
            the resulting queryset filtering takes allocated
            permissions into account using Django's permissions
            framework.

            Only returns records if user has dashboard permissions to
            do so. See `has_view_listboard_perms`.

            Limit records to those created by the current user if
            `has_view_only_my_listboard_perms` return True.
            See `has_view_only_my_listboard_perms`.

        Applies additional filter/exclude criteria.
        """

        queryset = self.listboard_model_cls.objects.none()
        if self.has_view_listboard_perms:
            q_exclude, opts_exclude = self.get_queryset_exclude_options(
                self.request, *self.args, **self.kwargs
            )
            q_filter, opts_filter = self.get_queryset_filter_options(
                self.request, *self.args, **self.kwargs
            )

            queryset = self.listboard_model_cls.objects.filter(
                q_filter, **opts_filter
            ).exclude(q_exclude, **opts_exclude)

            queryset = self.get_updated_queryset(queryset)

            ordering = self.get_ordering()
            if ordering:
                if isinstance(ordering, (str,)):
                    ordering = (ordering,)
                queryset = queryset.order_by(*ordering)
        return queryset

    def get_queryset_filter_options(self, request, *args, **kwargs) -> tuple[Q, dict]:
        """Returns filtering applied to every queryset"""
        options = dict(site_id__in=sites.get_site_ids_for_user(request=self.request))
        if self.has_view_only_my_listboard_perms:
            options.update(user_created=self.request.user.username)
        return Q(), options

    def get_queryset_exclude_options(self, request, *args, **kwargs) -> tuple[Q, dict]:
        """Returns exclude options applied to every queryset"""
        return Q(), {}

    def get_empty_queryset_message(self) -> str:
        return self.empty_queryset_message

    def get_updated_queryset(self, queryset):
        """Return the queryset for this view.

        Hook for a last chance to modify the queryset
        before ordering.
        """
        return queryset

    @property
    def has_view_listboard_perms(self):
        """Returns True if request.user has permissions to
        view the listboard.

        If False, `get_queryset` returns an empty queryset.
        """
        return self.request.user.has_perm(self.listboard_view_permission_codename)

    @property
    def has_view_only_my_listboard_perms(self):
        """Returns True if `request.user` ONLY has permissions to
        view records created by `request.user` on the listboard.
        """
        return self.request.user.has_perm(self.listboard_view_only_my_permission_codename)

    @property
    def has_listboard_model_perms(self):
        """Returns True if `request.user` has permissions to
        add/change the listboard model.

        Does not affect `get_queryset`.

        Used in templates.
        """
        app_label = self.listboard_model_cls._meta.label_lower.split(".")[0]
        model_name = self.listboard_model_cls._meta.label_lower.split(".")[1]
        return self.request.user.has_perms(
            [f"{app_label}.add_{model_name}", f"{app_label}.change_{model_name}"]
        )


class ListboardView(
    QueryStringViewMixin,
    UrlRequestContextMixin,
    SearchListboardMixin,
    BaseListboardView,
):
    urlconfig_getattr = "listboard_urls"

    @classmethod
    def get_urlname(cls):
        return cls.listboard_url



from django import forms
from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist

from edc_constants.constants import NO, YES
from edc_form_validators import FormValidator
from edc_utils import convert_php_dateformat
from edc_visit_tracking.constants import MISSED_VISIT
from edc_visit_tracking.utils import get_related_visit_model_cls

from .constants import LOST_TO_FOLLOWUP


class LossToFollowupFormValidator(FormValidator):
    def clean(self):
        self.check_if_last_visit_was_missed()
        self.required_if(YES, field="phone", field_required="phone_attempts")
        self.required_if(YES, field="home_visit", field_required="home_visit_detail")
        if (
            self.cleaned_data.get("phone_attempts") == 0
            and self.cleaned_data.get("home_visit") == NO
        ):
            raise forms.ValidationError(
                "No contact attempted. An attempt must be made to contact "
                "the patient by phone or home visit before declaring as lost "
                "to follow up."
            )
        self.validate_other_specify(
            field="loss_category", other_specify_field="loss_category_other"
        )

    def check_if_last_visit_was_missed(self):
        last_obj = (
            get_related_visit_model_cls()
            .objects.filter(
                appointment__subject_identifier=self.cleaned_data.get("subject_identifier"),
            )
            .last()
        )
        if last_obj.reason != MISSED_VISIT:
            raise forms.ValidationError(
                f"Wait! Last visit was not reported as `missed`. Got {last_obj}"
            )
        return True


class RequiresLtfuFormValidatorMixin:
    """Used in off schedule form or any form that
    needs to confirm the LTFU form was submitted
    first.
    """

    ltfu_model = None  # "inte_prn.losstofollowup"
    ltfu_date_field = None  # "ltfu_date"
    ltfu_reason = LOST_TO_FOLLOWUP

    @property
    def ltfu_model_cls(self):
        return django_apps.get_model(self.ltfu_model)

    def validate_ltfu(self):
        if self.ltfu_model and (self.cleaned_data.get("subject_identifier") or self.instance):
            subject_identifier = (
                self.cleaned_data.get("subject_identifier") or self.instance.subject_identifier
            )

            try:
                ltfu = django_apps.get_model(self.ltfu_model).objects.get(
                    subject_identifier=subject_identifier
                )
            except ObjectDoesNotExist:
                if (
                    self.cleaned_data.get(self.offschedule_reason_field)
                    and self.cleaned_data.get(self.offschedule_reason_field).name
                    == self.ltfu_reason
                ):
                    msg = (
                        "Patient is lost to followup, please complete "
                        f"`{self.ltfu_model_cls._meta.verbose_name}` "
                        "form first."
                    )
                    raise forms.ValidationError({self.offschedule_reason_field: msg})
            else:
                if self.cleaned_data.get(self.ltfu_date_field) and (
                    ltfu.ltfu_date != self.cleaned_data.get(self.ltfu_date_field)
                ):
                    expected = ltfu.ltfu_date.strftime(
                        convert_php_dateformat(settings.SHORT_DATE_FORMAT)
                    )
                    got = self.cleaned_data.get(self.ltfu_date_field).strftime(
                        convert_php_dateformat(settings.SHORT_DATE_FORMAT)
                    )
                    raise forms.ValidationError(
                        {
                            self.ltfu_date_field: (
                                "Date does not match "
                                f"`{self.ltfu_model_cls._meta.verbose_name}` "
                                f"form. Expected {expected}. Got {got}."
                            )
                        }
                    )



from .ae_initial_ae_model_mixin import AeInitialAeModelMixin
from .ae_initial_fields_model_mixin import AeInitialFieldsModelMixin
from .ae_initial_methods_model_mixin import AeInitialMethodsModelMixin
from .ae_initial_model_mixin import AeInitialModelMixin
from .ae_initial_sae_model_mixin import AeInitialSaeModelMixin
from .ae_initial_susar_model_mixin import AeInitialSusarModelMixin
from .ae_initial_tmg_model_mixin import AeInitialTmgModelMixin



# Generated by Django 4.2.7 on 2023-12-04 02:32

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_lab", "0029_alter_aliquot_options_alter_box_options_and_more"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="aliquot",
            name="edc_lab_ali_modifie_49b6da_idx",
        ),
        migrations.RemoveIndex(
            model_name="box",
            name="edc_lab_box_modifie_af5133_idx",
        ),
        migrations.RemoveIndex(
            model_name="boxitem",
            name="edc_lab_box_modifie_6a2a97_idx",
        ),
        migrations.RemoveIndex(
            model_name="boxtype",
            name="edc_lab_box_modifie_eb0207_idx",
        ),
        migrations.RemoveIndex(
            model_name="manifest",
            name="edc_lab_man_modifie_aae866_idx",
        ),
        migrations.RemoveIndex(
            model_name="manifestitem",
            name="edc_lab_man_modifie_113dc0_idx",
        ),
        migrations.RemoveIndex(
            model_name="order",
            name="edc_lab_ord_modifie_33b427_idx",
        ),
        migrations.RemoveIndex(
            model_name="result",
            name="edc_lab_res_modifie_edced5_idx",
        ),
        migrations.RemoveIndex(
            model_name="resultitem",
            name="edc_lab_res_modifie_d7258d_idx",
        ),
        migrations.RemoveIndex(
            model_name="shipper",
            name="edc_lab_shi_modifie_822b3b_idx",
        ),
        migrations.AddIndex(
            model_name="aliquot",
            index=models.Index(
                fields=["modified", "created"], name="edc_lab_ali_modifie_d3de00_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="box",
            index=models.Index(
                fields=["modified", "created"], name="edc_lab_box_modifie_58c481_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="boxitem",
            index=models.Index(
                fields=["modified", "created"], name="edc_lab_box_modifie_2ea786_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="boxtype",
            index=models.Index(
                fields=["modified", "created"], name="edc_lab_box_modifie_0dfa0d_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="manifest",
            index=models.Index(
                fields=["modified", "created"], name="edc_lab_man_modifie_771947_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="manifestitem",
            index=models.Index(
                fields=["modified", "created"], name="edc_lab_man_modifie_01a620_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="order",
            index=models.Index(
                fields=["modified", "created"], name="edc_lab_ord_modifie_8102e9_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="result",
            index=models.Index(
                fields=["modified", "created"], name="edc_lab_res_modifie_798e22_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="resultitem",
            index=models.Index(
                fields=["modified", "created"], name="edc_lab_res_modifie_8457dc_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="shipper",
            index=models.Index(
                fields=["modified", "created"], name="edc_lab_shi_modifie_73bada_idx"
            ),
        ),
    ]



# Generated by Django 3.2.11 on 2022-03-04 01:07

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0001_initial"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="medication",
            unique_together={("name", "display_name")},
        ),
    ]



from .autocomplete_admin import (
    UnblindingRequestorUserAdmin,
    UnblindingReviewerUserAdmin,
)
from .unblinding_request_admin import UnblindingRequestAdmin
from .unblinding_review_admin import UnblindingReviewAdmin



# Generated by Django 6.0 on 2025-03-04 16:19

from django.db import migrations

import edc_pharmacy.models.stock.stock_request


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0057_scanduplicates"),
    ]

    operations = [
        migrations.CreateModel(
            name="StockRequestProxy",
            fields=[],
            options={
                "verbose_name": "Stock Request",
                "verbose_name_plural": "Stock Requests",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("edc_pharmacy.stockrequest",),
            managers=[
                ("objects", edc_pharmacy.models.stock.stock_request.Manager()),
            ],
        ),
        migrations.AlterModelOptions(
            name="stockproxy",
            options={"verbose_name": "Stock", "verbose_name_plural": "Stock"},
        ),
    ]



from __future__ import annotations

from string import Formatter
from typing import TYPE_CHECKING

from django.apps import apps as django_apps

from edc_protocol.research_protocol_config import ResearchProtocolConfig

from .checkdigit_mixins import LuhnMixin
from .exceptions import IdentifierError

if TYPE_CHECKING:
    from django.contrib.sites.models import Site

    from .models import IdentifierModel


class IdentifierMissingTemplateValue(Exception):  # noqa: N818
    pass


class ResearchIdentifier:
    label: str = None  # e.g. subject_identifier, plot_identifier, etc
    identifier_type: str | None = None  # e.g. 'subject', 'infant', 'plot', a.k.a subject_type
    template: str | None = None
    padding: int = 5
    checkdigit = LuhnMixin()

    def __init__(
        self,
        identifier_type: str | None = None,
        template: str | None = None,
        device_id: str | None = None,
        protocol_number: str | None = None,
        site: Site | None = None,
        requesting_model: str | None = None,
        identifier: str | None = None,
    ) -> None:
        self._identifier = None
        self._sequence_number = None
        self.requesting_model = requesting_model
        if not self.requesting_model:
            raise IdentifierError("Invalid requesting_model. Got None")
        self.identifier_type = identifier_type or self.identifier_type
        if not self.identifier_type:
            raise IdentifierError("Invalid identifier_type. Got None")
        self.template = template or self.template
        app_config = django_apps.get_app_config("edc_device")
        self.device_id = device_id or app_config.device_id
        self.protocol_number = protocol_number or ResearchProtocolConfig().protocol_number
        self.site = site or django_apps.get_model("sites.site").objects.get_current()
        if identifier:
            # load an existing identifier
            self.identifier_model = self.identifier_model_cls.objects.get(
                identifier=identifier
            )
            self._identifier = self.identifier_model.identifier
            self.subject_type = self.identifier_model.subject_type
            self.site = self.identifier_model.site
        self.get_identifier()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.label})"

    def __str__(self) -> str:
        return self.identifier

    @property
    def identifier_model_cls(self) -> IdentifierModel:
        return django_apps.get_model("edc_identifier.identifiermodel")

    @property
    def identifier(self) -> str:
        """Returns a new and unique identifier and updates
        the IdentifierModel.
        """
        if not self._identifier:
            self.pre_identifier()
            self._identifier = self.template.format(**self.template_opts)
            check_digit = self.checkdigit.calculate_checkdigit(
                "".join(self._identifier.split("-"))
            )
            self._identifier = f"{self._identifier}-{check_digit}"
            self.identifier_model = self.identifier_model_cls.objects.create(
                **self.identifier_options
            )
            self.post_identifier()
        return self._identifier

    def get_identifier(self) -> str:
        return self.identifier

    @property
    def identifier_options(self) -> dict:
        return dict(
            name=self.label,
            sequence_number=self.sequence_number,
            identifier=self._identifier,
            protocol_number=self.protocol_number,
            device_id=self.device_id,
            model=self.requesting_model,
            site=self.site,
            identifier_type=self.identifier_type,
        )

    def pre_identifier(self) -> None:
        pass

    def post_identifier(self) -> None:
        pass

    @property
    def template_opts(self) -> dict:
        """Returns the template key/values, if a key from the template
        does not exist raises an exception.
        """
        template_opts: dict = {}
        formatter = Formatter()
        keys = [opt[1] for opt in formatter.parse(self.template) if opt[1] not in ["sequence"]]
        template_opts.update(sequence=str(self.sequence_number).rjust(self.padding, "0"))
        for key in keys:
            try:
                value = getattr(self, key)
            except AttributeError as e:
                raise IdentifierMissingTemplateValue(
                    f"Required option not provided. Got '{key}'."
                ) from e
            else:
                if value:
                    template_opts.update({key: value})
                else:
                    raise IdentifierMissingTemplateValue(
                        f"Required option cannot be None. Got '{key}'."
                    )
        return template_opts

    @property
    def site_id(self) -> str:
        return str(self.site.pk)

    @property
    def sequence_number(self) -> int:
        """Returns the next sequence number to use."""
        if self._sequence_number is None:
            if identifier_model := (
                self.identifier_model_cls.objects.filter(
                    name=self.label, device_id=self.device_id, site=self.site
                )
                .order_by("-sequence_number")
                .first()
            ):
                self._sequence_number = identifier_model.sequence_number + 1
            else:
                self._sequence_number = 1
        return self._sequence_number



from __future__ import annotations

from secrets import choice
from typing import TYPE_CHECKING, Any

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.utils import timezone

from .utils import convert_to_human_readable

if TYPE_CHECKING:
    from edc_identifier.models import IdentifierModel


class DuplicateIdentifierError(Exception):
    pass


class IdentifierError(Exception):
    pass


class SimpleIdentifier:
    random_string_length: int = 5
    template: str = "{device_id}{random_string}"
    identifier_prefix: str = None

    def __init__(
        self,
        template: str | None = None,
        random_string_length: int | None = None,
        identifier_prefix: str | None = None,
        device_id: str | None = None,
    ) -> None:
        self._identifier: str | None = None
        self.template = template or self.template
        self.random_string_length = random_string_length or self.random_string_length
        self.device_id = device_id or django_apps.get_app_config("edc_device").device_id
        self.identifier_prefix = identifier_prefix or self.identifier_prefix

    def __str__(self) -> str:
        return self.identifier

    @property
    def identifier(self) -> str:
        if not self._identifier:
            self._identifier = self.template.format(
                device_id=self.device_id, random_string=self.random_string
            )
            if self.identifier_prefix:
                self._identifier = f"{self.identifier_prefix}{self._identifier}"
        return self._identifier

    @property
    def random_string(self) -> str:
        return "".join(
            [
                choice("ABCDEFGHKMNPRTUVWXYZ2346789")  # nosec B311
                for _ in range(self.random_string_length)
            ]
        )


class SimpleTimestampIdentifier(SimpleIdentifier):
    timestamp_format: str = "%y%m%d%H%M%S%f"
    timestamp_length: int = 14

    @property
    def identifier(self) -> str:
        if not self._identifier:
            self._identifier = self.template.format(
                device_id=self.device_id,
                timestamp=timezone.localtime().strftime(self.timestamp_format)[
                    : self.timestamp_length
                ],
                random_string=self.random_string,
            )
            if self.identifier_prefix:
                self._identifier = f"{self.identifier_prefix}{self._identifier}"
        return self._identifier


class SimpleSequentialIdentifier:
    prefix: str | None = None

    def __init__(self):
        sequence: int = int(timezone.now().timestamp())
        random_number: int = choice(range(1000, 9999))  # nosec B311
        sequence: str = f"{sequence}{random_number}"
        chk: int = int(sequence) % 11
        self.identifier: str = f"{self.prefix or ''}{sequence}{chk}"

    def __str__(self) -> str:
        return self.identifier


class SimpleUniqueIdentifier:
    """Usage:

    class ManifestIdentifier(Identifier):
        random_string_length = 9
        identifier_attr = "manifest_identifier"
        template = "M{device_id}{random_string}"
    """

    random_string_length: int = 5
    identifier_type: str = "simple_identifier"
    identifier_attr: str = "identifier"
    model: str = "edc_identifier.identifiermodel"
    template: str = "{device_id}{random_string}"
    identifier_prefix: str | None = None
    identifier_cls = SimpleIdentifier
    make_human_readable: bool | None = None

    def __init__(
        self,
        model: str = None,
        identifier_attr: str | None = None,
        identifier_type: str | None = None,
        identifier_prefix: str | None = None,
        make_human_readable: bool | None = None,
        linked_identifier: str | None = None,
        protocol_number: str | None = None,
        source_model: str | None = None,
        subject_identifier: str | None = None,
        name: str | None = None,
        site: Any | None = None,
        site_id: str | None = None,
    ):
        self._identifier: str | None = None
        self.site_id = site_id
        self.name = name or ""
        self.model = model or self.model
        self.linked_identifier = linked_identifier
        self.protocol_number = protocol_number
        self.source_model = source_model
        self.subject_identifier = subject_identifier
        self.identifier_attr = identifier_attr or self.identifier_attr
        self.identifier_type = identifier_type or self.identifier_type
        self.identifier_prefix = identifier_prefix or self.identifier_prefix
        if self.identifier_prefix and len(self.identifier_prefix) != 2:
            raise IdentifierError(
                f"Expected identifier_prefix of length=2. Got {len(identifier_prefix)}"
            )
        self.make_human_readable = make_human_readable or self.make_human_readable
        self.device_id = django_apps.get_app_config("edc_device").device_id
        if site:
            self.site_id = site.id
        elif site_id:
            self.site_id = site_id

    def __str__(self):
        return self.identifier

    @property
    def identifier(self) -> str:
        if not self._identifier:
            tries = 0
            while not self._identifier:
                tries += 1
                self._identifier = self._get_new_identifier()
                if self.make_human_readable:
                    self._identifier = convert_to_human_readable(self._identifier)
                try:
                    self.model_cls.objects.get(
                        identifier_type=self.identifier_type,
                        **{self.identifier_attr: self._identifier},
                    )
                except ObjectDoesNotExist:
                    opts = dict(
                        sequence_number=1,
                        linked_identifier=self.linked_identifier,
                        protocol_number=self.protocol_number,
                        model=self.source_model,
                        subject_identifier=self.subject_identifier,
                        name=self.name,
                    )
                    opts.update({self.identifier_attr: self._identifier})
                    self.update_identifier_model(**opts)
                if tries > 100:
                    raise DuplicateIdentifierError(
                        "Unable prepare a unique identifier, "
                        "all are taken. Increase the length of the random string"
                    )
        return self._identifier

    def _get_new_identifier(self) -> str:
        """Returns a new identifier."""
        identifier = self.identifier_cls(
            template=self.template,
            identifier_prefix=self.identifier_prefix,
            random_string_length=self.random_string_length,
            device_id=self.device_id,
        )
        return identifier.identifier

    @property
    def model_cls(self) -> type[models.Model]:
        return django_apps.get_model(self.model)

    def update_identifier_model(self, **kwargs) -> bool | IdentifierModel:
        """Attempts to update identifier_model and returns True (or instance)
        if successful else False if identifier already exists.
        """
        opts = dict(
            identifier=self.identifier,
            identifier_type=self.identifier_type,
            identifier_prefix=self.identifier_prefix,
            device_id=self.device_id,
            site_id=self.site_id,
        )
        opts.update(**kwargs)
        try:
            self.model_cls.objects.get(identifier=self.identifier)
        except ObjectDoesNotExist:
            opts = {k: v for k, v in opts.items() if v is not None}
            return self.model_cls.objects.create(**opts)
        return False



from edc_navbar import NavbarItem

dm_navbar_item = NavbarItem(
    name="data_manager_home",
    title="Data Management",
    label="DM",
    codename="edc_data_manager.nav_data_manager_section",
    url_name="edc_data_manager:home_url",
)



from uuid import uuid4

from django.db import models
from django.utils import timezone

from ...stock import ContainerType
from ..box import Box


class ContainerModelMixin(models.Model):
    container_identifier = models.CharField(max_length=36, default=uuid4, unique=True)

    container_datetime = models.DateTimeField(default=timezone.now)

    name = models.CharField(max_length=25, unique=True)

    container_type = models.ForeignKey(ContainerType, on_delete=models.PROTECT)

    box = models.ForeignKey(Box, on_delete=models.PROTECT, null=True)

    description = models.TextField(default="")

    contains_uniquely_identifiable_items = models.BooleanField(default=True)

    def __str__(self):
        return f"{self.container_type.name} {self.name}"

    def save(self, *args, **kwargs):
        if not self.name:
            self.name = self.container_identifier
        super().save(*args, **kwargs)

    class Meta:
        abstract = True
        verbose_name = "Item"
        verbose_name_plural = "Items"



import sys

from django.apps import AppConfig as DjangoAppConfig
from django.core.management.color import color_style

from .device import Device
from .device_permission import device_permissions

style = color_style()


class AppConfig(DjangoAppConfig):
    device_cls = Device

    def __init__(self, app_name, app_module):
        self._device_id = None
        super().__init__(app_name, app_module)

    name = "edc_device"
    verbose_name = "Edc Device"
    include_in_administration_section = True

    messages_written = False

    device_id = None
    device_role = None

    central_server_id = "99"
    middleman_id_list = ["95"]
    node_server_id_list = ["98"]

    def ready(self):
        from .signals import check_device_on_pre_save  # noqa

        device = Device(
            device_id=self.device_id,
            device_role=self.device_role,
            central_server_id=self.central_server_id,
            middlemen=self.middleman_id_list,
            nodes=self.node_server_id_list,
        )
        # set app_config instance attrs including device_id, device_role
        for k, v in device.__dict__.items():
            setattr(self, k, v)

        if not self.messages_written:
            self.messages_written = True
            sys.stdout.write(f"Loading {self.verbose_name} ...\n")
            sys.stdout.write(
                f"  * device id is '{self.device_id}'.\n"
                f"  * device role is '{self.device_role}'.\n"
            )
            for index, device_permission in enumerate(device_permissions):
                if index == 0:
                    sys.stdout.write("  * device permissions exist for:\n")
                sys.stdout.write(f"    - {device_permission}\n")
            sys.stdout.write(f" Done loading {self.verbose_name}.\n")



from edc_constants.constants import OTHER, YES
from edc_dx import get_diagnosis_labels
from edc_dx.form_validators import DiagnosisFormValidatorMixin
from edc_visit_schedule.utils import raise_if_baseline


class ClinicalReviewFollowupFormValidatorMixin(DiagnosisFormValidatorMixin):
    def _clean(self):
        raise_if_baseline(self.cleaned_data.get("subject_visit"))
        for prefix, label in get_diagnosis_labels().items():
            cond = prefix.lower()
            self.applicable_if_not_diagnosed(
                prefix=cond,
                field_applicable=f"{cond}_test",
                label=label,
            )
            self.required_if(YES, field=f"{cond}_test", field_required=f"{cond}_test_date")
            self.m2m_required_if(YES, field=f"{cond}_test", m2m_field=f"{cond}_reason")
            self.m2m_other_specify(
                OTHER,
                m2m_field=f"{cond}_reason",
                field_other=f"{cond}_reason_other",
            )
            self.applicable_if(YES, field=f"{cond}_test", field_applicable=f"{cond}_dx")
        super()._clean()



from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from ..constants import ON_SCHEDULE
from ..model_mixins import OffScheduleModelMixin, OnScheduleModelMixin
from ..site_visit_schedules import SiteVisitScheduleError, site_visit_schedules


@receiver(post_save, weak=False, dispatch_uid="offschedule_model_on_post_save")
def offschedule_model_on_post_save(sender, instance, raw, update_fields, **kwargs):
    if not raw and not update_fields and isinstance(instance, (OffScheduleModelMixin,)):
        _, schedule = site_visit_schedules.get_by_offschedule_model(instance._meta.label_lower)
        schedule.take_off_schedule(instance.subject_identifier, instance.offschedule_datetime)


@receiver(post_delete, weak=False, dispatch_uid="offschedule_model_on_post_delete")
def offschedule_model_on_post_delete(instance, **kwargs):
    if isinstance(instance, (OffScheduleModelMixin,)):
        try:
            _, schedule = site_visit_schedules.get_by_offschedule_model(
                instance._meta.label_lower
            )
        except SiteVisitScheduleError:
            pass
        else:
            if schedule.offschedule_model == instance._meta.label_lower:
                history_obj = schedule.history_model_cls.objects.get(
                    subject_identifier=instance.subject_identifier,
                    onschedule_model=schedule.onschedule_model,
                )
                history_obj.offschedule_datetime = None
                history_obj.schedule_status = ON_SCHEDULE
                history_obj.save()
                onschedule_model_obj = schedule.onschedule_model_cls.objects.get(
                    subject_identifier=instance.subject_identifier
                )
                onschedule_model_obj.save()


@receiver(post_delete, weak=False, dispatch_uid="onschedule_model_on_post_delete")
def onschedule_model_on_post_delete(instance, **kwargs):
    if isinstance(instance, (OnScheduleModelMixin,)):
        try:
            _, schedule = site_visit_schedules.get_by_offschedule_model(
                instance._meta.label_lower
            )
        except SiteVisitScheduleError:
            pass
        else:
            if schedule.onschedule_model == instance._meta.label_lower:
                schedule.history_model_cls.objects.filter(
                    subject_identifier=instance.subject_identifier
                ).delete()


@receiver(post_save, weak=False, dispatch_uid="put_subject_on_schedule_on_post_save")
def put_subject_on_schedule_on_post_save(sender, instance, raw, created, **kwargs):
    if not raw:
        try:
            instance.put_subject_on_schedule_on_post_save(created)
        except AttributeError as e:
            if "put_subject_on_schedule_on_post_save" not in str(e):
                raise



from django.db import models

from edc_reportable.units import (
    MICROMOLES_PER_LITER,
    MICROMOLES_PER_LITER_DISPLAY,
    MILLIGRAMS_PER_DECILITER,
    MILLIMOLES_PER_LITER,
    MILLIMOLES_PER_LITER_DISPLAY,
)

from ..model_mixin_factories import reportable_result_model_mixin_factory


class CreatinineModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="creatinine",
        verbose_name="Creatinine",
        units_choices=(
            (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
            (MICROMOLES_PER_LITER, MICROMOLES_PER_LITER_DISPLAY),
        ),
    ),
    models.Model,
):
    class Meta:
        abstract = True


class UreaModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="urea",
        verbose_name="Urea (BUN)",
        units_choices=((MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER_DISPLAY),),
    ),
    models.Model,
):
    class Meta:
        abstract = True


class UricAcidModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="uric_acid",
        verbose_name="Uric Acid",
        decimal_places=4,
        max_digits=10,
        units_choices=(
            (MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER_DISPLAY),
            (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
        ),
    ),
    models.Model,
):
    class Meta:
        abstract = True



from __future__ import annotations

from django.apps import apps as django_apps


def get_models(app_config, exclude_models: list[str] | None):
    exclude_models = exclude_models or []
    return [m for m in app_config.get_models() if m._meta.label_lower not in exclude_models]


def get_app_codenames(
    *crud_apps: str,
    list_app: str | None = None,
    autocomplete_models: list[str] | None = None,
    permissions: list[str] | None = None,
    exclude_models: list[str] | tuple[str, ...] | None = None,
    include_import_export: bool | None = None,
) -> list[str]:
    """Prepares and returns an ordered list of codenames for the
    common edc project apps to be used in `auth_objects`.

    Reads default_permissions from model.Meta or permissions
    to build a list of codenames for each model in the app_config.

    For example, in `auths.py`:

        clinic_codenames = get_app_codenames(
            "meta_prn", "meta_subject", "meta_consent", list_app="meta_lists"
        )
    """

    codenames: list[str] = []
    autocomplete_models = autocomplete_models or []
    exclude_models = exclude_models or []
    try:
        app_config = django_apps.get_app_config(list_app)
    except LookupError:
        pass
    else:
        for model_cls in get_models(app_config, exclude_models):
            codenames.append(f"{app_config.name}.view_{model_cls._meta.model_name}")
    for name in crud_apps:
        try:
            app_config = django_apps.get_app_config(name)
        except LookupError:
            pass
        else:
            for model_cls in get_models(app_config, exclude_models):
                codenames.extend(
                    get_codename(
                        app_config.name,
                        model_cls,
                        autocomplete_models,
                        override_permissions=permissions,
                        include_import_export=include_import_export,
                    )
                )
    codenames.sort()
    return codenames


def get_codename(
    app_name,
    model_cls,
    autocomplete_models: list[str] | None = None,
    override_permissions: list[str] | None = None,
    include_import_export: bool | None = None,
) -> list[str]:
    codenames = []
    autocomplete_models = autocomplete_models or []
    label_lower: str = model_cls._meta.label_lower
    model_name: str = model_cls._meta.model_name
    if "historical" in label_lower or label_lower in autocomplete_models:
        codenames.append(f"{app_name}.view_{model_name}")
    else:
        permissions = override_permissions or model_cls._meta.default_permissions
        if not include_import_export:
            permissions = [
                x for x in permissions if x not in ["export", "import", "viewallsites"]
            ]
        prefixes = [f"{s}_" for s in permissions]
        for prefix in prefixes:
            codenames.append(f"{app_name}.{prefix}{model_name}")
    return codenames



from .aliquot_types import bc, disposable, fbc, pl, serum, sputum, urine, wb
from .constants import PACKED, SHIPPED
from .identifiers import (
    AliquotIdentifier,
    AliquotIdentifierCountError,
    AliquotIdentifierLengthError,
    BoxIdentifier,
    ManifestIdentifier,
    RequisitionIdentifier,
)
from .lab import (
    AliquotCreator,
    AliquotType,
    LabProfile,
    Manifest,
    Process,
    ProcessingProfile,
    RequisitionPanel,
    RequisitionPanelGroup,
    Specimen,
    SpecimenProcessor,
)
from .site_labs import site_labs



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "edc_pylabels"
    verbose_name = "Edc Labels"
    include_in_administration_section = True



# Generated by Django 2.1.2 on 2018-10-24 23:50

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_notification", "0002_update_notifications"),
        ("edc_auth", "0006_userprofile_notifications"),
    ]

    operations = [
        migrations.RemoveField(model_name="userprofile", name="notifications"),
        migrations.AddField(
            model_name="userprofile",
            name="email_notifications",
            field=models.ManyToManyField(
                blank=True,
                limit_choices_to={"enabled": True},
                related_name="email_notifications",
                to="edc_notification.Notification",
            ),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="mobile",
            field=models.CharField(
                blank=True,
                help_text="e.g. +1234567890",
                max_length=25,
                null=True,
                validators=[django.core.validators.RegexValidator(regex="^\\+\\d+")],
            ),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="sms_notifications",
            field=models.ManyToManyField(
                blank=True,
                limit_choices_to={"enabled": True},
                related_name="sms_notifications",
                to="edc_notification.Notification",
            ),
        ),
    ]



from edc_auth.constants import CLINICIAN_ROLE, CLINICIAN_SUPER_ROLE

from ..auth_objects import PHARMACIST_ROLE, PHARMACY_SUPER_ROLE, SITE_PHARMACIST_ROLE


class AuthsViewMixin:

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            roles=[obj.name for obj in self.request.user.userprofile.roles.all()],
            SITE_PHARMACIST_ROLE=SITE_PHARMACIST_ROLE,
            PHARMACIST_ROLE=PHARMACIST_ROLE,
            PHARMACY_SUPER_ROLE=PHARMACY_SUPER_ROLE,
            CLINICIAN_ROLE=CLINICIAN_ROLE,
            CLINICIAN_SUPER_ROLE=CLINICIAN_SUPER_ROLE,
        )
        return context



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models.deletion import PROTECT
from django.utils import timezone

from edc_constants.constants import CANCELLED, NEW
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_notification.model_mixins import NotificationModelMixin
from edc_sites.managers import CurrentSiteManager as BaseCurrentSiteManager
from edc_sites.model_mixins import SiteModelMixin

from ..choices import ACTION_STATUS, PRIORITY
from ..exceptions import ActionItemStatusError, SubjectDoesNotExist
from ..identifiers import ActionIdentifier
from ..site_action_items import site_action_items
from .action_type import ActionType

if TYPE_CHECKING:
    from edc_crf.model_mixins import CrfModelMixin

    from ..action import Action

    class PrnModel(BaseUuidModel):
        subject_identifier: str
        ...

    class CrfModel(CrfModelMixin): ...


RELATED_ACTION_ITEM_DOES_NOT_EXIST = (
    "Related ActionItem does not exist. Got {action_identifier}."
)
PARENT_ACTION_ITEM_DOES_NOT_EXIST = (
    "Parent ActionItem does not exist. Got {action_identifier}."
)
INVALID_ACTION_ITEM_STATUS = (
    "Invalid action item status. Reference model exists. "
    "Got `{status}`. Perhaps catch this in the form"
)

INVALID_SUBJECT_IDENTIFIER = (
    "Invalid subject identifier. Attempt to create {class_name} failed. "
    "Subject does not exist in '{subject_identifier_model}'. "
    "Got '{subject_identifier}'."
)


class CurrentSiteManager(BaseCurrentSiteManager):
    use_in_migrations = True

    def get_by_natural_key(self, action_identifier):
        return self.get(action_identifier=action_identifier)


class ActionItemManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, action_identifier):
        return self.get(action_identifier=action_identifier)


class ActionItem(
    NonUniqueSubjectIdentifierFieldMixin,
    SiteModelMixin,
    NotificationModelMixin,
    BaseUuidModel,
):
    subject_identifier_model = "edc_registration.registeredsubject"

    action_identifier = models.CharField(max_length=50, unique=True)

    report_datetime = models.DateTimeField(default=timezone.now)

    action_type = models.ForeignKey(
        ActionType, on_delete=PROTECT, related_name="action_type", verbose_name="Action"
    )

    linked_to_reference = models.BooleanField(
        default=False,
        editable=False,
        help_text=(
            "True if this action is linked to it's reference_model."
            "Initially False if this action is created before reference_model."
            "Always True when reference_model creates the action."
            'Set to True when reference_model is created and "links" to this action.'
            "(Note: reference_model looks for actions where "
            "linked_to_reference is False before attempting to "
            "create a new ActionItem)."
        ),
    )

    # related_reference_model = models.CharField(max_length=100, null=True, editable=False)

    related_action_identifier = models.CharField(
        max_length=50,
        blank=True,
        default="",
        help_text=(
            "May be left blank. e.g. action identifier from source model that opened the item."
        ),
    )

    parent_action_identifier = models.CharField(
        max_length=50,
        blank=True,
        default="",
        help_text=(
            "May be left blank. e.g. action identifier from "
            "reference model that opened the item (parent)."
        ),
    )

    priority = models.CharField(
        max_length=25,
        choices=PRIORITY,
        blank=True,
        default="",
        help_text="Leave blank to use default for this action type.",
    )

    parent_action_item = models.ForeignKey(
        "self",
        on_delete=PROTECT,
        related_name="+",
        null=True,
        blank=True,
        editable=False,
    )

    related_action_item = models.ForeignKey(
        "self",
        on_delete=PROTECT,
        related_name="+",
        null=True,
        blank=True,
        editable=False,
    )

    status = models.CharField(max_length=25, default=NEW, choices=ACTION_STATUS)

    instructions = models.TextField(
        blank=True, default="", help_text="populated by action class"
    )

    auto_created = models.BooleanField(default=False)

    auto_created_comment = models.CharField(max_length=25, blank=True, default="")

    objects = ActionItemManager()

    on_site = CurrentSiteManager()

    history = HistoricalRecords()

    def __str__(self) -> str:
        return (
            f"{self.action_type.display_name} {self.action_identifier[-9:]} "
            f"for {self.subject_identifier} ({self.get_status_display()})"
        )

    def save(self, *args, **kwargs):
        """See also signals and action_cls."""
        if not self.id:
            # a new persisted action item always has
            # a unique action identifier
            self.action_identifier = (
                self.action_identifier
                or ActionIdentifier(
                    subject_identifier=self.subject_identifier,
                    site_id=self.site_id,
                    source_model=self._meta.label_lower,
                    name=self.action_type.name,
                ).identifier
            )
            # subject_identifier
            subject_identifier_model_cls = django_apps.get_model(self.subject_identifier_model)
            try:
                subject_identifier_model_cls.objects.get(
                    subject_identifier=self.subject_identifier
                )
            except ObjectDoesNotExist as e:
                raise SubjectDoesNotExist(
                    INVALID_SUBJECT_IDENTIFIER.format(
                        class_name=self.__class__.__name__,
                        subject_identifier_model=self.subject_identifier_model,
                        subject_identifier=self.subject_identifier,
                    )
                ) from e
            self.priority = self.priority or self.action_type.priority
        elif (
            self.status in [NEW, CANCELLED]
            and django_apps.get_model(self.reference_model)
            .objects.filter(action_identifier=self.action_identifier)
            .exists()
        ):
            raise ActionItemStatusError(
                INVALID_ACTION_ITEM_STATUS.format(status=self.get_status_display())
            )
        super().save(*args, **kwargs)

    def natural_key(self) -> tuple[str]:
        return (self.action_identifier,)

    @property
    def last_updated(self) -> datetime | str:
        return None if self.status == NEW else self.modified

    @property
    def user_last_updated(self) -> str | None:
        return None if self.status == NEW else self.user_modified or self.user_created

    @property
    def action_cls(self) -> type[Action]:
        """Returns the action_cls."""
        return site_action_items.get(self.action_type.name)

    @property
    def action(self) -> Action:
        """Returns the instantiated action_cls."""
        return self.action_cls(
            subject_identifier=self.subject_identifier,
            action_identifier=self.action_identifier,
            readonly=True,
        )

    @property
    def reference_model(self) -> str:
        return self.action_type.reference_model

    @property
    def reference_model_cls(self) -> type[PrnModel | CrfModel]:
        return django_apps.get_model(self.action_type.reference_model)

    @property
    def reference_obj(self) -> PrnModel | CrfModel:
        return self.reference_model_cls.objects.get(action_identifier=self.action_identifier)

    @property
    def parent_reference_obj(self) -> PrnModel | CrfModel:
        if not self.parent_action_item:
            raise ObjectDoesNotExist(
                PARENT_ACTION_ITEM_DOES_NOT_EXIST.format(
                    action_identifier=self.action_identifier
                )
            )
        return self.parent_action_item.reference_obj

    @property
    def related_reference_model(self) -> str:
        return self.action_type.related_reference_model

    @property
    def related_reference_obj(self) -> PrnModel | CrfModel:
        if not self.related_action_item:
            raise ObjectDoesNotExist(
                RELATED_ACTION_ITEM_DOES_NOT_EXIST.format(
                    action_identifier=self.action_identifier
                )
            )
        return self.related_action_item.reference_obj

    @property
    def identifier(self) -> str:
        """Returns a shortened action identifier."""
        return self.action_identifier[-9:]

    @property
    def display_name(self) -> str:
        return self.action_type.display_name

    @property
    def label_color(self) -> str:
        return self.action_cls.color_style

    class Meta(
        BaseUuidModel.Meta,
        NonUniqueSubjectIdentifierFieldMixin.Meta,
    ):
        verbose_name = "Action Item"
        verbose_name_plural = "Action Items"
        indexes = (
            *BaseUuidModel.Meta.indexes,
            *NonUniqueSubjectIdentifierFieldMixin.Meta.indexes,
            *[
                models.Index(
                    fields=[
                        "subject_identifier",
                        "action_identifier",
                        "action_type",
                        "site_id",
                        "status",
                        "report_datetime",
                    ],
                ),
            ],
        )



# Generated by Django 5.1.2 on 2024-11-18 11:28

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0041_alter_dispenseitem_dispense_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="historicaldispensinghistory",
            name="history_user",
        ),
        migrations.RemoveField(
            model_name="historicaldispensinghistory",
            name="rx_refill",
        ),
        migrations.RemoveField(
            model_name="historicalreturnhistory",
            name="history_user",
        ),
        migrations.RemoveField(
            model_name="historicalreturnhistory",
            name="rx_refill",
        ),
        migrations.AlterUniqueTogether(
            name="returnhistory",
            unique_together=None,
        ),
        migrations.RemoveField(
            model_name="returnhistory",
            name="rx_refill",
        ),
        migrations.DeleteModel(
            name="DispensingHistory",
        ),
        migrations.DeleteModel(
            name="HistoricalDispensingHistory",
        ),
        migrations.DeleteModel(
            name="HistoricalReturnHistory",
        ),
        migrations.DeleteModel(
            name="ReturnHistory",
        ),
    ]



from .consent_modelform_mixin import ConsentModelFormMixin, ReviewFieldsModelFormMixin
from .requires_consent_modelform_mixin import RequiresConsentModelFormMixin

__all__ = [
    "ConsentModelFormMixin",
    "RequiresConsentModelFormMixin",
    "ReviewFieldsModelFormMixin",
]



import warnings

from .form_validator_mixins import FormValidatorMixin  # noqa

warnings.warn(
    "This path is deprecated in favor of form_validator_mixins.",
    DeprecationWarning,
    stacklevel=2,
)



# Generated by Django 2.1.5 on 2019-02-01 02:46

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_lab", "0017_auto_20181108_0353")]

    operations = [
        migrations.AlterField(
            model_name="historicalaliquot",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
        migrations.AlterField(
            model_name="historicalbox",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="consignee",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_lab.Consignee",
                verbose_name="Consignee",
            ),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="shipper",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_lab.Shipper",
                verbose_name="Shipper/Exporter",
            ),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
        migrations.AlterField(
            model_name="historicalresult",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
        migrations.AlterField(
            model_name="historicalresultitem",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
    ]



from django.contrib.auth.decorators import login_required
from django.utils.decorators import method_decorator
from django.views.generic.base import TemplateView

from edc_dashboard.view_mixins import EdcViewMixin


class HomeView(EdcViewMixin, TemplateView):
    template_name = "edc_identifier/home.html"

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)
